# Orca
#
# Copyright 2013-2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-public-methods
# pylint: disable=too-many-locals
# pylint: disable=too-many-lines

"""Provides an Orca-controlled caret for text content."""

from __future__ import annotations

import functools
from typing import TYPE_CHECKING

from . import (
    caret_navigator_command_definitions,
    command_manager,
    dbus_service,
    debug,
    focus_manager,
    gsettings_registry,
    guilabels,
    input_event,
    input_event_manager,
    messages,
    presentation_manager,
    say_all_presenter,
    script_manager,
)
from .ax_object import AXObject
from .ax_text import AXText
from .ax_utilities import AXUtilities
from .ax_utilities_text import CaretSetReason
from .extension import Extension

if TYPE_CHECKING:
    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi

    from .command import Command
    from .scripts import default


@gsettings_registry.get_registry().gsettings_schema(
    "org.gnome.Orca.CaretNavigation",
    name="caret-navigation",
)
class CaretNavigator(Extension):
    """Implements the caret navigation support available to scripts."""

    _SCHEMA = "caret-navigation"
    KEY_ENABLED = "enabled"
    KEY_TRIGGERS_FOCUS_MODE = "triggers-focus-mode"
    KEY_LAYOUT_MODE = "layout-mode"

    def _get_setting(self, key: str, default: bool) -> bool:
        """Returns the dconf value for key, or default if not in dconf."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            key,
            "b",
            default=default,
        )

    GROUP_LABEL = guilabels.KB_GROUP_CARET_NAVIGATION

    def __init__(self) -> None:
        # To make it possible for focus mode to suspend this navigation without
        # changing the user's preferred setting.
        self._suspended: bool = False
        self._last_input_event: input_event.InputEvent | None = None
        self._enabled_for_script: dict[default.Script, bool] = {}
        super().__init__()

    @staticmethod
    def navigation_command(func):
        """Decorator that logs the command, then dispatches to it."""

        @functools.wraps(func)
        def wrapper(self, script, event=None, notify_user=True) -> bool:
            tokens = [
                "CARET NAVIGATOR:",
                func,
                "\nScript:",
                script,
                "\nEvent:",
                event,
                "\nnotify_user:",
                notify_user,
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return func(self, script, event, notify_user)

        return wrapper

    def _get_commands(self) -> list[Command]:
        return caret_navigator_command_definitions.get_commands(self)

    def _is_active_script(self, script):
        active_script = script_manager.get_manager().get_active_script()
        if active_script == script:
            return True

        tokens = ["CARET NAVIGATOR:", script, "is not the active script", active_script]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return False

    @gsettings_registry.get_registry().gsetting(
        key=KEY_ENABLED,
        schema="caret-navigation",
        gtype="b",
        default=True,
        summary="Enable caret navigation",
        migration_key="caretNavigationEnabled",
    )
    @dbus_service.getter
    def get_is_enabled(self) -> bool:
        """Returns whether caret navigation is enabled."""

        return self._get_setting(self.KEY_ENABLED, True)

    @dbus_service.setter
    def set_is_enabled(self, value: bool) -> bool:
        """Sets whether caret navigation is enabled."""

        if self.get_is_enabled() == value:
            msg = f"CARET NAVIGATOR: Enabled already {value}. Refreshing command group."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            command_manager.get_manager().set_group_enabled(
                guilabels.KB_GROUP_CARET_NAVIGATION,
                value,
            )
            return True

        msg = f"CARET NAVIGATOR: Setting enabled to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(self._SCHEMA, self.KEY_ENABLED, value)

        self._last_input_event = None
        command_manager.get_manager().set_group_enabled(guilabels.KB_GROUP_CARET_NAVIGATION, value)

        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_TRIGGERS_FOCUS_MODE,
        schema="caret-navigation",
        gtype="b",
        default=False,
        summary="Caret navigation triggers focus mode",
        migration_key="caretNavTriggersFocusMode",
    )
    @dbus_service.getter
    def get_triggers_focus_mode(self) -> bool:
        """Returns whether caret navigation triggers focus mode."""

        return self._get_setting(self.KEY_TRIGGERS_FOCUS_MODE, False)

    @dbus_service.setter
    def set_triggers_focus_mode(self, value: bool) -> bool:
        """Sets whether caret navigation triggers focus mode."""

        if self.get_triggers_focus_mode() == value:
            return True

        msg = f"CARET NAVIGATOR: Setting triggers focus mode to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_TRIGGERS_FOCUS_MODE,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_LAYOUT_MODE,
        schema="caret-navigation",
        gtype="b",
        default=True,
        summary="Use document layout mode",
        migration_key="layoutMode",
    )
    @dbus_service.getter
    def get_layout_mode(self) -> bool:
        """Returns whether layout mode is enabled."""

        return self._get_setting(self.KEY_LAYOUT_MODE, True)

    @dbus_service.setter
    def set_layout_mode(self, value: bool) -> bool:
        """Sets whether layout mode is enabled."""

        if self.get_layout_mode() == value:
            return True

        msg = f"CARET NAVIGATOR: Setting layout mode to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_LAYOUT_MODE, value
        )
        return True

    @dbus_service.command
    def toggle_layout_mode(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Switches between object mode and layout mode for line presentation."""

        tokens = [
            "CARET NAVIGATOR: toggle_layout_mode. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        layout_mode = not self.get_layout_mode()
        if notify_user:
            if layout_mode:
                presentation_manager.get_manager().present_message(messages.MODE_LAYOUT)
            else:
                presentation_manager.get_manager().present_message(messages.MODE_OBJECT)
        self.set_layout_mode(layout_mode)
        return True

    def get_enabled_for_script(self, script: default.Script) -> bool:
        """Returns the current caret-navigator enabled state associated with script."""

        enabled = self._enabled_for_script.get(script, False)
        tokens = ["CARET NAVIGATOR: Enabled state for", script, f"is {enabled}"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return enabled

    def set_enabled_for_script(self, script: default.Script, enabled: bool) -> None:
        """Sets the current caret-navigator enabled state associated with script."""

        tokens = ["CARET NAVIGATOR: Setting enabled state for", script, f"to {enabled}"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        self._enabled_for_script[script] = enabled

        if not (script and self._is_active_script(script)):
            return

        # Use the per-script state combined with the user's preference to determine
        # whether commands should be active, without overwriting the preference.
        effective = enabled and self.get_is_enabled()
        command_manager.get_manager().set_group_enabled(
            guilabels.KB_GROUP_CARET_NAVIGATION,
            effective,
        )

    def last_input_event_was_navigation_command(self) -> bool:
        """Returns true if the last input event was a navigation command."""

        if self._last_input_event is None:
            return False

        manager = input_event_manager.get_manager()
        result = manager.last_event_equals_or_is_release_for_event(self._last_input_event)
        if self._last_input_event is not None:
            string = self._last_input_event.as_single_line_string()
        else:
            string = "None"

        msg = f"CARET NAVIGATOR: Last navigation event ({string}) is last input event: {result}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    def last_command_prevents_focus_mode(self) -> bool:
        """Returns True if the last command was navigation but the setting disallows focus mode."""

        if not self.last_input_event_was_navigation_command():
            return False

        return not self.get_triggers_focus_mode()

    @dbus_service.command
    def toggle_enabled(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Toggles caret navigation."""

        tokens = [
            "CARET NAVIGATOR: toggle_enabled. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        enabled = not command_manager.get_manager().is_group_enabled(
            guilabels.KB_GROUP_CARET_NAVIGATION,
        )
        if enabled:
            string = messages.CARET_CONTROL_ORCA
        else:
            string = messages.CARET_CONTROL_APP
            script.utilities.clear_caret_context()

        if notify_user:
            presentation_manager.get_manager().present_message(string)

        self.set_is_enabled(enabled)
        return True

    def suspend_commands(self, script: default.Script, suspended: bool, reason: str = "") -> None:
        """Suspends caret navigation independent of the enabled setting."""

        if not (script and self._is_active_script(script)):
            return

        msg = f"CARET NAVIGATOR: Commands suspended: {suspended}"
        if reason:
            msg += f": {reason}"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        self._suspended = suspended
        command_manager.get_manager().set_group_suspended(
            guilabels.KB_GROUP_CARET_NAVIGATION,
            suspended,
        )

    def _get_root_object(
        self,
        script: default.Script,
        obj: Atspi.Accessible | None = None,
    ) -> Atspi.Accessible | None:
        """Returns the object which should be treated as the root/container for navigation."""

        root = script.utilities.active_document()
        if root is None:
            if obj is None:
                obj, _offset = script.utilities.get_caret_context()
            if AXObject.supports_text(obj):
                root = obj

        tokens = ["CARET NAVIGATOR: Root is", root]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return root

    def _is_navigable_object(
        self,
        script: default.Script,
        obj: Atspi.Accessible,
        root: Atspi.Accessible | None = None,
    ) -> bool:
        """Returns True if obj is a valid location for navigation."""

        # There's a small, theoretical possibility that we can creep out of the logical container,
        # but until that happens, this check is the most performant.
        if AXObject.supports_text(obj):
            return True

        if root is None:
            root = self._get_root_object(script)

        if root is None:
            return False

        return AXUtilities.is_ancestor(obj, root, True)

    def _line_contains_context(
        self,
        line: list[tuple[Atspi.Accessible, int, int, str]],
        context: tuple[Atspi.Accessible, int],
    ) -> bool:
        """Returns True if line contains the (obj, offset) context."""

        for entry in line:
            line_obj, start, end = entry[0], entry[1], entry[2]
            if line_obj == context[0] and start <= context[1] <= end:
                return True

        return False

    def _get_embedded_document_frame(self, script: default.Script) -> Atspi.Accessible | None:
        """Returns the embedded document frame that confines file-boundary navigation, if any."""

        obj, _offset = script.utilities.get_caret_context()
        if obj is None:
            return None
        return AXUtilities.get_embedded_document_frame_for_object(obj)

    def _get_start_of_file(self, script: default.Script) -> tuple[Atspi.Accessible | None, int]:
        """Returns the start of the file as (obj, offset)."""

        frame = self._get_embedded_document_frame(script)
        root = frame if frame is not None else self._get_root_object(script)
        obj, offset = script.utilities.first_context(root, 0)
        if obj is None:
            return None, -1

        while obj:
            prev_obj, prev_offset = script.utilities.previous_context(obj, offset, restrict_to=root)
            if prev_obj is None or (prev_obj, prev_offset) == (obj, offset):
                break
            # The web context walkers ignore restrict_to, so enforce the frame boundary here.
            if frame is not None and not AXUtilities.is_ancestor(prev_obj, frame, True):
                break
            obj, offset = prev_obj, prev_offset

        return obj, offset

    def _get_end_of_file(self, script: default.Script) -> tuple[Atspi.Accessible | None, int]:
        """Returns the end of the file as (obj, offset)."""

        frame = self._get_embedded_document_frame(script)
        root = frame if frame is not None else self._get_root_object(script)
        obj = AXUtilities.find_deepest_descendant(root)
        if obj is None:
            return None, -1

        # Chromium includes static text leaf nodes which we ignore; use the navigable parent.
        if not AXUtilities.is_web_element(obj):
            obj = AXObject.get_parent(obj)

        offset = max(0, AXText.get_character_count(obj) - 1)
        while obj:
            next_obj, next_offset = script.utilities.next_context(obj, offset, restrict_to=root)
            if next_obj is None or (next_obj, next_offset) == (obj, offset):
                break
            # The web context walkers ignore restrict_to, so enforce the frame boundary here.
            if frame is not None and not AXUtilities.is_ancestor(next_obj, frame, True):
                break
            obj, offset = next_obj, next_offset

        return obj, offset

    @dbus_service.command
    @navigation_command
    def next_character(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the next character."""

        obj, offset = script.utilities.next_context()
        if not self._is_navigable_object(script, obj):
            return False

        string, char_start, char_end = AXText.get_character_at_offset(
            obj, offset, ensure_whole_characters=True
        )
        if string and offset > char_start:
            offset = char_end

        self._last_input_event = event
        presentation_manager.get_manager().interrupt_presentation()
        script.utilities.set_caret_position(obj, offset, reason=CaretSetReason.CARET_NAVIGATION)
        focus_manager.get_manager().emit_region_changed(
            obj,
            start_offset=offset,
            mode=focus_manager.CARET_NAVIGATOR,
        )
        if not notify_user:
            return True

        script.update_braille(obj, offset=offset)
        script.say_character(obj, offset)
        return True

    @dbus_service.command
    @navigation_command
    def previous_character(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the previous character."""

        obj, offset = script.utilities.previous_context()
        if not self._is_navigable_object(script, obj):
            return False

        string, char_start, _char_end = AXText.get_character_at_offset(
            obj, offset, ensure_whole_characters=True
        )
        if string and offset > char_start:
            offset = char_start

        self._last_input_event = event
        presentation_manager.get_manager().interrupt_presentation()
        script.utilities.set_caret_position(obj, offset, reason=CaretSetReason.CARET_NAVIGATION)
        focus_manager.get_manager().emit_region_changed(
            obj,
            start_offset=offset,
            mode=focus_manager.CARET_NAVIGATOR,
        )
        if not notify_user:
            return True

        script.update_braille(obj, offset=offset)
        script.say_character(obj, offset)
        return True

    @dbus_service.command
    @navigation_command
    def next_word(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the next word."""

        obj, offset = script.utilities.next_context(skip_space=True)
        if obj is None:
            return False

        contents = script.utilities.get_word_contents_at_offset(obj, offset)
        if not contents:
            return False

        # If the "word" to the right consists of the content of the last word in an embedded
        # object followed by the space of the parent object, the normal space-adjustment we
        # do will cause us to set the caret to the offset with the embedded child and then
        # present the first word in that child.
        if len(contents) > 1 and contents[-1][3].isspace():
            msg = "CARET NAVIGATOR: Adjusting next word contents to eliminate trailing space."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            contents = contents[:-1]

        obj, start, end, string = contents[-1]
        if not self._is_navigable_object(script, obj):
            return False

        # Strip trailing whitespace so a paragraph break does not cause the word to be skipped.
        end = start + len(string.rstrip())

        self._last_input_event = event
        presentation_manager.get_manager().interrupt_presentation()
        script.utilities.set_caret_position(obj, end, reason=CaretSetReason.CARET_NAVIGATION)
        focus_manager.get_manager().emit_region_changed(
            obj,
            start,
            end,
            focus_manager.CARET_NAVIGATOR,
        )
        if not notify_user:
            return True

        script.update_braille(obj, offset=end)
        script.say_word(obj, end)
        return True

    @dbus_service.command
    @navigation_command
    def previous_word(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the previous word."""

        obj, offset = script.utilities.previous_context(skip_space=True)
        if obj is None:
            return False

        contents = script.utilities.get_word_contents_at_offset(obj, offset)
        if not contents:
            return False

        obj, start, end, _string = contents[0]
        if not self._is_navigable_object(script, obj):
            return False

        self._last_input_event = event
        presentation_manager.get_manager().interrupt_presentation()
        script.utilities.set_caret_position(obj, start, reason=CaretSetReason.CARET_NAVIGATION)
        focus_manager.get_manager().emit_region_changed(
            obj,
            start,
            end,
            focus_manager.CARET_NAVIGATOR,
        )

        if not notify_user:
            return True

        script.update_braille(obj, offset=start)
        script.say_word(obj, start)
        return True

    @dbus_service.command
    @navigation_command
    def next_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the next line."""

        if (
            focus_manager.get_manager().in_say_all()
            and say_all_presenter.get_presenter().get_rewind_and_fast_forward_enabled()
        ):
            msg = "CARET NAVIGATOR: In say all and rewind/fast-forward is enabled"
            debug.print_message(debug.LEVEL_INFO, msg)
            return True

        obj, offset = script.utilities.get_caret_context()
        if obj is None:
            return False

        line = script.utilities.get_line_contents_at_offset(obj, offset)
        if not (line and line[0]):
            return False

        contents = script.utilities.get_next_line_contents()
        if not contents:
            last_obj, last_offset = self._get_end_of_file(script)
            if self._line_contains_context(line, (last_obj, last_offset)):
                msg = "CARET NAVIGATOR: At end of document; cannot move to next line."
                debug.print_message(debug.LEVEL_INFO, msg)
                contents = line

        if not contents:
            return False

        if not self._is_navigable_object(script, obj):
            return False

        self._last_input_event = event
        presentation_manager.get_manager().interrupt_presentation()

        if line != contents:
            obj, offset, end, _string = contents[0]
        else:
            obj, offset, end, _string = contents[-1]

        script.utilities.set_caret_position(obj, offset, reason=CaretSetReason.CARET_NAVIGATION)
        focus_manager.get_manager().emit_region_changed(
            obj,
            offset,
            end,
            focus_manager.CARET_NAVIGATOR,
        )

        if notify_user:
            # Setting the last object on the current line as priorObj
            # prevents re-announcing context.
            presenter = presentation_manager.get_manager()
            presenter.speak_contents(contents, prior_obj=line[-1][0])
            presenter.display_contents(contents)
        return True

    @dbus_service.command
    @navigation_command
    def previous_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the previous line."""

        if (
            focus_manager.get_manager().in_say_all()
            and say_all_presenter.get_presenter().get_rewind_and_fast_forward_enabled()
        ):
            msg = "CARET NAVIGATOR: In say all and rewind/fast-forward is enabled"
            debug.print_message(debug.LEVEL_INFO, msg)
            return True

        obj, offset = script.utilities.get_caret_context()
        if obj is None:
            return False

        line = script.utilities.get_line_contents_at_offset(obj, offset)
        if not (line and line[0]):
            return False

        contents = script.utilities.get_previous_line_contents(obj, offset)
        if not contents:
            first_obj, first_offset = self._get_start_of_file(script)
            if self._line_contains_context(line, (first_obj, first_offset)):
                msg = "CARET NAVIGATOR: At start of document; cannot move to previous line."
                debug.print_message(debug.LEVEL_INFO, msg)
                contents = line

        if not contents:
            return False

        obj, start, end, _string = contents[0]
        if not self._is_navigable_object(script, obj):
            return False

        self._last_input_event = event
        presentation_manager.get_manager().interrupt_presentation()
        script.utilities.set_caret_position(obj, start, reason=CaretSetReason.CARET_NAVIGATION)
        focus_manager.get_manager().emit_region_changed(
            obj,
            start,
            end,
            focus_manager.CARET_NAVIGATOR,
        )

        if notify_user:
            # Setting the first object on the current line as priorObj
            # prevents re-announcing context.
            presenter = presentation_manager.get_manager()
            presenter.speak_contents(contents, prior_obj=line[0][0])
            presenter.display_contents(contents)
        return True

    @dbus_service.command
    @navigation_command
    def start_of_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the start of the line."""

        obj, offset = script.utilities.get_caret_context()
        line = script.utilities.get_line_contents_at_offset(obj, offset)
        if not (line and line[0]):
            return False

        self._last_input_event = event
        obj, start, end, _string = line[0]
        presentation_manager.get_manager().interrupt_presentation()
        script.utilities.set_caret_position(obj, start, reason=CaretSetReason.CARET_NAVIGATION)
        focus_manager.get_manager().emit_region_changed(
            obj,
            start,
            end,
            focus_manager.CARET_NAVIGATOR,
        )

        if not notify_user:
            return True

        script.say_character(obj, start)
        presentation_manager.get_manager().display_contents(line)
        return True

    @dbus_service.command
    @navigation_command
    def end_of_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the end of the line."""

        obj, offset = script.utilities.get_caret_context()
        line = script.utilities.get_line_contents_at_offset(obj, offset)
        if not (line and line[0]):
            return False

        obj, start, end, string = line[-1]
        if string.strip() and string[-1].isspace():
            end -= 1

        self._last_input_event = event
        presentation_manager.get_manager().interrupt_presentation()
        script.utilities.set_caret_position(obj, end, reason=CaretSetReason.CARET_NAVIGATION)
        focus_manager.get_manager().emit_region_changed(
            obj,
            start,
            end,
            focus_manager.CARET_NAVIGATOR,
        )

        if not notify_user:
            return True

        script.say_character(obj, end)
        presentation_manager.get_manager().display_contents(line)
        return True

    @dbus_service.command
    @navigation_command
    def start_of_file(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the start of the file."""

        obj, start = self._get_start_of_file(script)
        if obj is None:
            return False

        contents = script.utilities.get_line_contents_at_offset(obj, start)
        if not contents:
            return False

        self._last_input_event = event
        obj, start, end, _string = contents[0]
        presentation_manager.get_manager().interrupt_presentation()
        script.utilities.set_caret_position(obj, start, reason=CaretSetReason.CARET_NAVIGATION)
        focus_manager.get_manager().emit_region_changed(
            obj,
            start,
            end,
            focus_manager.CARET_NAVIGATOR,
        )

        if not notify_user:
            return True

        presenter = presentation_manager.get_manager()
        presenter.speak_contents(contents)
        presenter.display_contents(contents)
        return True

    @dbus_service.command
    @navigation_command
    def end_of_file(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the end of the file."""

        obj, end = self._get_end_of_file(script)
        if obj is None:
            return False

        contents = script.utilities.get_line_contents_at_offset(obj, end)
        if not contents:
            return False

        self._last_input_event = event
        obj, start, end, _string = contents[-1]
        presentation_manager.get_manager().interrupt_presentation()
        script.utilities.set_caret_position(obj, end, reason=CaretSetReason.CARET_NAVIGATION)
        focus_manager.get_manager().emit_region_changed(
            obj,
            start,
            end,
            focus_manager.CARET_NAVIGATOR,
        )
        if not notify_user:
            return True

        presenter = presentation_manager.get_manager()
        presenter.speak_contents(contents)
        presenter.display_contents(contents)
        return True


_navigator = CaretNavigator()


def get_navigator() -> CaretNavigator:
    """Returns the Caret Navigator."""

    return _navigator



# Orca
#
# Copyright 2024 Igalia, S.L.
# Copyright 2024 GNOME Foundation Inc.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-lines
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-branches
# pylint: disable=too-many-statements

"""Utilities for accessible events."""

from __future__ import annotations

import enum
import re
import time
from difflib import SequenceMatcher
from typing import TYPE_CHECKING

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

from . import ax_cache_manager, debug, focus_manager
from .ax_object import AXObject
from .ax_selection import AXSelection
from .ax_text import AXText
from .ax_utilities_collection import AXUtilitiesCollection
from .ax_utilities_component import AXUtilitiesComponent
from .ax_utilities_object import AXUtilitiesObject
from .ax_utilities_relation import AXUtilitiesRelation
from .ax_utilities_role import AXUtilitiesRole
from .ax_utilities_state import AXUtilitiesState
from .ax_utilities_text import AXUtilitiesText, CaretSetReason
from .ax_value import AXValue

if TYPE_CHECKING:
    from .input_event_manager import InputEventManager


class TextEventReason(enum.Enum):
    """Enum representing the reason for an object:text- event."""

    UNKNOWN = enum.auto()
    AUTO_DELETION_PRESENTABLE = enum.auto()
    AUTO_DELETION_UNPRESENTABLE = enum.auto()
    AUTO_INSERTION_PRESENTABLE = enum.auto()
    AUTO_INSERTION_UNPRESENTABLE = enum.auto()
    AUTO_SELECTION = enum.auto()
    AUTO_UNSELECTION = enum.auto()
    BACKSPACE = enum.auto()
    BRAILLE_PANNING = enum.auto()
    CHILDREN_CHANGE = enum.auto()
    CUT = enum.auto()
    DELETE = enum.auto()
    FOCUS_CHANGE = enum.auto()
    LIVE_REGION_UPDATE = enum.auto()
    MOUSE_MIDDLE_BUTTON = enum.auto()
    MOUSE_PRIMARY_BUTTON = enum.auto()
    NAVIGATION_BY_CHARACTER = enum.auto()
    NAVIGATION_BY_LINE = enum.auto()
    NAVIGATION_BY_PARAGRAPH = enum.auto()
    NAVIGATION_BY_PAGE = enum.auto()
    NAVIGATION_BY_WORD = enum.auto()
    NAVIGATION_TO_FILE_BOUNDARY = enum.auto()
    NAVIGATION_TO_LINE_BOUNDARY = enum.auto()
    PAGE_SWITCH = enum.auto()
    PASTE = enum.auto()
    REDO = enum.auto()
    SAY_ALL = enum.auto()
    SEARCH_PRESENTABLE = enum.auto()
    SEARCH_UNPRESENTABLE = enum.auto()
    SELECT_ALL = enum.auto()
    SELECTED_TEXT_DELETION = enum.auto()
    SELECTED_TEXT_INSERTION = enum.auto()
    SELECTED_TEXT_RESTORATION = enum.auto()
    SELECTION_BY_CHARACTER = enum.auto()
    SELECTION_BY_LINE = enum.auto()
    SELECTION_BY_PARAGRAPH = enum.auto()
    SELECTION_BY_PAGE = enum.auto()
    SELECTION_BY_WORD = enum.auto()
    SELECTION_TO_FILE_BOUNDARY = enum.auto()
    SELECTION_TO_LINE_BOUNDARY = enum.auto()
    SPIN_BUTTON_VALUE_CHANGE = enum.auto()
    SYSTEM_UPDATE = enum.auto()
    TERMINAL_LINE_NAVIGATION_REPAINT = enum.auto()
    TYPING = enum.auto()
    TYPING_ECHOABLE = enum.auto()
    UI_UPDATE = enum.auto()
    UNDO = enum.auto()
    UNSPECIFIED_COMMAND = enum.auto()
    UNSPECIFIED_NAVIGATION = enum.auto()
    UNSPECIFIED_SELECTION = enum.auto()


class _AXUtilitiesEventCache:
    """Provides event-specific access to manager-backed cached values."""

    LAST_KNOWN_DESCRIPTION = "AXUtilitiesEvent.last-known-description"
    LAST_KNOWN_NAME = "AXUtilitiesEvent.last-known-name"
    LAST_KNOWN_CHECKED = "AXUtilitiesEvent.last-known-checked"
    LAST_KNOWN_EXPANDED = "AXUtilitiesEvent.last-known-expanded"
    LAST_KNOWN_INDETERMINATE = "AXUtilitiesEvent.last-known-indeterminate"
    LAST_KNOWN_INVALID_ENTRY = "AXUtilitiesEvent.last-known-invalid-entry"
    LAST_KNOWN_PRESSED = "AXUtilitiesEvent.last-known-pressed"
    LAST_KNOWN_SELECTED = "AXUtilitiesEvent.last-known-selected"
    LAST_KNOWN_VALUE = "AXUtilitiesEvent.last-known-value"
    LAST_KNOWN_VALUE_TEXT = "AXUtilitiesEvent.last-known-value-text"
    IGNORE_NAME_CHANGES_FOR = "AXUtilitiesEvent.ignore-name-changes-for"
    TEXT_EVENT_REASON = "AXUtilitiesEvent.text-event-reason"

    _NAMESPACES = (
        LAST_KNOWN_DESCRIPTION,
        LAST_KNOWN_NAME,
        LAST_KNOWN_CHECKED,
        LAST_KNOWN_EXPANDED,
        LAST_KNOWN_INDETERMINATE,
        LAST_KNOWN_INVALID_ENTRY,
        LAST_KNOWN_PRESSED,
        LAST_KNOWN_SELECTED,
        LAST_KNOWN_VALUE,
        LAST_KNOWN_VALUE_TEXT,
        IGNORE_NAME_CHANGES_FOR,
        TEXT_EVENT_REASON,
    )

    def __init__(self) -> None:
        self._manager = ax_cache_manager.get_manager()
        for namespace in self._NAMESPACES:
            self._manager.register_cache(
                self,
                namespace,
                lifetime=ax_cache_manager.Lifetime.PROCESS,
            )
        self._caches = {
            namespace: self._manager.get_cache(self, namespace) for namespace in self._NAMESPACES
        }

    def get_description(self, obj: Atspi.Accessible) -> str | None:
        """Returns the cached description for obj."""

        cache = self._caches.get(self.LAST_KNOWN_DESCRIPTION)
        if cache is None:
            return None

        return cache.get(ax_cache_manager.get_object_key(obj), None)

    def set_description(self, obj: Atspi.Accessible, description: str) -> None:
        """Stores the description for obj."""

        cache = self._caches.get(self.LAST_KNOWN_DESCRIPTION)
        if cache is not None:
            cache.put(ax_cache_manager.get_object_key(obj), description)

    def get_name(self, obj: Atspi.Accessible) -> str | None:
        """Returns the cached name for obj."""

        cache = self._caches.get(self.LAST_KNOWN_NAME)
        if cache is None:
            return None

        return cache.get(ax_cache_manager.get_object_key(obj), None)

    def set_name(self, obj: Atspi.Accessible, name: str) -> None:
        """Stores the name for obj."""

        cache = self._caches.get(self.LAST_KNOWN_NAME)
        if cache is not None:
            cache.put(ax_cache_manager.get_object_key(obj), name)

    def get_state(self, namespace: str, obj: Atspi.Accessible) -> bool | None:
        """Returns the cached state for obj."""

        cache = self._caches.get(namespace)
        if cache is None:
            return None

        return cache.get(ax_cache_manager.get_object_key(obj), None)

    def set_state(self, namespace: str, obj: Atspi.Accessible, state: bool) -> None:
        """Stores the state for obj."""

        cache = self._caches.get(namespace)
        if cache is not None:
            cache.put(ax_cache_manager.get_object_key(obj), state)

    def get_value(self, obj: Atspi.Accessible) -> float | None:
        """Returns the cached value for obj."""

        cache = self._caches.get(self.LAST_KNOWN_VALUE)
        if cache is None:
            return None

        return cache.get(ax_cache_manager.get_object_key(obj), None)

    def set_value(self, obj: Atspi.Accessible, value: float) -> None:
        """Stores the value for obj."""

        cache = self._caches.get(self.LAST_KNOWN_VALUE)
        if cache is not None:
            cache.put(ax_cache_manager.get_object_key(obj), value)

    def get_value_text(self, obj: Atspi.Accessible) -> str | None:
        """Returns the cached value text for obj."""

        cache = self._caches.get(self.LAST_KNOWN_VALUE_TEXT)
        if cache is None:
            return None

        return cache.get(ax_cache_manager.get_object_key(obj), None)

    def set_value_text(self, obj: Atspi.Accessible, value_text: str) -> None:
        """Stores the value text for obj."""

        cache = self._caches.get(self.LAST_KNOWN_VALUE_TEXT)
        if cache is not None:
            cache.put(ax_cache_manager.get_object_key(obj), value_text)

    def get_text_event_reason(self, event: Atspi.Event) -> TextEventReason | None:
        """Returns the cached reason for event."""

        cache = self._caches.get(self.TEXT_EVENT_REASON)
        if cache is None:
            return None

        return cache.get(event, None)

    def set_text_event_reason(self, event: Atspi.Event, reason: TextEventReason) -> None:
        """Stores the reason for event."""

        cache = self._caches.get(self.TEXT_EVENT_REASON)
        if cache is not None:
            cache.put(event, reason)

    def is_ignoring_name_changes_for(self, obj: Atspi.Accessible) -> bool:
        """Returns True if name changes should be ignored for obj."""

        cache = self._caches.get(self.IGNORE_NAME_CHANGES_FOR)
        if cache is None:
            return False

        return cache.get(ax_cache_manager.get_object_key(obj), False)

    def ignore_name_changes_for(self, obj: Atspi.Accessible) -> None:
        """Stores that name changes should be ignored for obj."""

        cache = self._caches.get(self.IGNORE_NAME_CHANGES_FOR)
        if cache is not None:
            cache.put(ax_cache_manager.get_object_key(obj), True)


class AXUtilitiesEvent:
    """Utilities for accessible events."""

    # How recent a caret set must be for its resulting event to be attributed to it.
    CARET_SET_EVENT_WINDOW_SECONDS = 1.0
    TERMINAL_REPAINT_EVENT_WINDOW_SECONDS = 1.0

    _CACHE = _AXUtilitiesEventCache()
    _LAST_TERMINAL_LINE_NAVIGATION_DELETION: tuple[object, float, str] | None = None
    _LAST_TERMINAL_LINE_NAVIGATION_REPAINT_LINE: tuple[object, float, str] | None = None

    @staticmethod
    def _save_terminal_line_navigation_deletion(event: Atspi.Event) -> None:
        AXUtilitiesEvent._LAST_TERMINAL_LINE_NAVIGATION_DELETION = (
            ax_cache_manager.get_object_key(event.source),
            time.monotonic(),
            event.any_data,
        )
        AXUtilitiesEvent._LAST_TERMINAL_LINE_NAVIGATION_REPAINT_LINE = None

    @staticmethod
    def _save_terminal_line_navigation_repaint_line(event: Atspi.Event) -> None:
        AXUtilitiesEvent._LAST_TERMINAL_LINE_NAVIGATION_REPAINT_LINE = (
            ax_cache_manager.get_object_key(event.source),
            time.monotonic(),
            AXText.get_line_at_offset(event.source)[0],
        )

    @staticmethod
    def _repeats_terminal_line_navigation_repaint_line(event: Atspi.Event) -> bool:
        prior = AXUtilitiesEvent._LAST_TERMINAL_LINE_NAVIGATION_REPAINT_LINE
        if prior is None:
            return False

        obj_key, timestamp, line = prior
        if obj_key != ax_cache_manager.get_object_key(event.source):
            return False

        if time.monotonic() - timestamp > AXUtilitiesEvent.TERMINAL_REPAINT_EVENT_WINDOW_SECONDS:
            return False

        return bool(line.strip()) and event.any_data.strip() == line.strip()

    @staticmethod
    def _is_terminal_line_navigation_repaint(event: Atspi.Event) -> bool:
        prior = AXUtilitiesEvent._LAST_TERMINAL_LINE_NAVIGATION_DELETION
        if prior is None:
            return False

        obj_key, timestamp, text = prior
        if obj_key != ax_cache_manager.get_object_key(event.source):
            return False

        if time.monotonic() - timestamp > AXUtilitiesEvent.TERMINAL_REPAINT_EVENT_WINDOW_SECONDS:
            return False

        deleted_lines = text.split("\n")
        inserted_lines = event.any_data.split("\n")
        if min(len(deleted_lines), len(inserted_lines)) < 4:
            return False

        # Terminal repaint text can have clipped edge lines. Compare the deleted
        # lines without their edges with the inserted lines shifted one row in either
        # direction. In tiny terminals, an expiring TUI status line can be another
        # edge row, leaving only the inner two shifted rows to compare.
        return (
            deleted_lines[1:-1] == inserted_lines[2:]
            or deleted_lines[2:] == inserted_lines[1:-1]
            or (
                min(len(deleted_lines), len(inserted_lines)) >= 5
                and (
                    deleted_lines[1:-2] == inserted_lines[2:-1]
                    or deleted_lines[2:-1] == inserted_lines[1:-2]
                )
            )
        )

    @staticmethod
    def _strings_are_redundant(str1: str | None, str2: str | None, threshold: float = 0.85) -> bool:
        """Returns True if str2 is redundant to str1 based on the similarity threshold."""

        if not (str1 and str2):
            return False

        similarity = round(SequenceMatcher(None, str1.lower(), str2.lower()).ratio(), 2)
        msg = (
            f"AXUtilitiesEvent: Similarity between '{str1}', '{str2}': {similarity} "
            f"(threshold: {threshold})"
        )
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return similarity >= threshold

    @staticmethod
    def save_object_info_for_events(obj: Atspi.Accessible) -> None:
        """Saves properties, states, etc. of obj for later use in event processing."""

        if obj is None:
            return

        state_set = AXObject.get_state_set(obj)
        AXUtilitiesEvent._CACHE.set_description(obj, AXObject.get_description(obj))
        AXUtilitiesEvent._CACHE.set_name(obj, AXObject.get_name(obj))
        AXUtilitiesEvent._CACHE.set_state(
            AXUtilitiesEvent._CACHE.LAST_KNOWN_CHECKED,
            obj,
            AXUtilitiesState.is_checked(obj, state_set),
        )
        AXUtilitiesEvent._CACHE.set_state(
            AXUtilitiesEvent._CACHE.LAST_KNOWN_EXPANDED,
            obj,
            AXUtilitiesState.is_expanded(obj, state_set),
        )
        AXUtilitiesEvent._CACHE.set_state(
            AXUtilitiesEvent._CACHE.LAST_KNOWN_INDETERMINATE,
            obj,
            AXUtilitiesState.is_indeterminate(obj, state_set),
        )
        AXUtilitiesEvent._CACHE.set_state(
            AXUtilitiesEvent._CACHE.LAST_KNOWN_PRESSED,
            obj,
            AXUtilitiesState.is_pressed(obj, state_set),
        )
        AXUtilitiesEvent._CACHE.set_state(
            AXUtilitiesEvent._CACHE.LAST_KNOWN_SELECTED,
            obj,
            AXUtilitiesState.is_selected(obj, state_set),
        )

        window = focus_manager.get_manager().get_active_window()
        AXUtilitiesEvent._CACHE.set_name(window, AXObject.get_name(window))
        AXUtilitiesEvent._CACHE.set_description(window, AXObject.get_description(window))

    @staticmethod
    def get_text_event_reason(event: Atspi.Event) -> TextEventReason:
        """Returns the TextEventReason for the given event."""

        reason = AXUtilitiesEvent._CACHE.get_text_event_reason(event)
        if reason is not None:
            tokens = ["AXUtilitiesEvent: Cached reason for", event, f": {reason}"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return reason

        reason = TextEventReason.UNKNOWN
        if event.type.startswith("object:text-changed:insert"):
            reason = AXUtilitiesEvent._get_text_insertion_event_reason(event)
        elif event.type.startswith("object:text-caret-moved"):
            reason = AXUtilitiesEvent._get_caret_moved_event_reason(event)
        elif event.type.startswith("object:text-changed:delete"):
            reason = AXUtilitiesEvent._get_text_deletion_event_reason(event)
        elif event.type.startswith("object:text-selection-changed"):
            reason = AXUtilitiesEvent._get_text_selection_changed_event_reason(event)
        else:
            raise ValueError(f"Unexpected event type: {event.type}")

        AXUtilitiesEvent._CACHE.set_text_event_reason(event, reason)
        tokens = ["AXUtilitiesEvent: Reason for", event, f": {reason}"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return reason

    @staticmethod
    def _is_terminal_being_flat_reviewed(obj: Atspi.Accessible) -> bool:
        """Returns True if obj is a terminal currently being flat-reviewed."""

        if not AXUtilitiesRole.is_terminal(obj):
            return False
        mode, acc = focus_manager.get_manager().get_active_mode_and_object_of_interest()
        return mode == focus_manager.FLAT_REVIEW and obj == acc

    @staticmethod
    def _is_terminal_escape_sequence(string: str) -> bool:
        """Returns True if string is an escape sequence the terminal displayed as literal text."""

        return bool(re.fullmatch(r"\s*ESC\[[\d;]*[A-Za-z~]?\s*", string))

    @staticmethod
    def _is_spin_button_descendant(obj: Atspi.Accessible) -> bool:
        """Returns True if obj is a spin button or descends from one."""

        if AXUtilitiesRole.is_spin_button(obj):
            return True
        return AXUtilitiesObject.find_ancestor(obj, AXUtilitiesRole.is_spin_button) is not None

    @staticmethod
    def _did_line_change(obj: Atspi.Accessible) -> bool:
        """Returns True if the last known line in obj differs from the current line."""

        last_obj, last_offset = focus_manager.get_manager().get_last_cursor_position()
        if obj != last_obj:
            return False

        return not AXUtilitiesText.offset_is_on_current_line(obj, last_offset)

    @staticmethod
    def _get_obj_type_reason_or_none(
        event: Atspi.Event,
        mgr: InputEventManager,
    ) -> TextEventReason | None:
        """Returns the reason if obj-type or event-type dictates one, else None to continue."""

        obj = event.source
        if AXUtilitiesRole.is_live_region(obj):
            return TextEventReason.LIVE_REGION_UPDATE
        if AXObject.get_role(obj) in AXUtilitiesRole.get_text_ui_roles():
            return TextEventReason.UI_UPDATE
        if event.type.endswith("system"):
            return TextEventReason.SYSTEM_UPDATE
        if mgr.last_event_was_page_switch():
            return TextEventReason.PAGE_SWITCH
        return None

    @staticmethod
    def _is_children_change(event: Atspi.Event) -> bool:
        """Returns True if the event's text is nothing but embedded object characters."""

        return bool(event.any_data) and not event.any_data.replace("\ufffc", "")

    @staticmethod
    def _get_non_editable_reason(mgr: InputEventManager) -> TextEventReason:
        """Returns the reason for an insert/delete in a non-editable, non-terminal object."""

        if mgr.last_event_was_command():
            return TextEventReason.UNSPECIFIED_COMMAND
        return TextEventReason.UNKNOWN

    @staticmethod
    def _get_selection_navigation_reason(mgr: InputEventManager) -> TextEventReason:
        """Returns the selection-granularity reason based on the last input event."""

        if mgr.last_event_was_line_navigation():
            reason = TextEventReason.SELECTION_BY_LINE
        elif mgr.last_event_was_word_navigation():
            reason = TextEventReason.SELECTION_BY_WORD
        elif mgr.last_event_was_character_navigation():
            reason = TextEventReason.SELECTION_BY_CHARACTER
        elif mgr.last_event_was_page_navigation():
            reason = TextEventReason.SELECTION_BY_PAGE
        elif mgr.last_event_was_line_boundary_navigation():
            reason = TextEventReason.SELECTION_TO_LINE_BOUNDARY
        elif mgr.last_event_was_file_boundary_navigation():
            reason = TextEventReason.SELECTION_TO_FILE_BOUNDARY
        else:
            reason = TextEventReason.UNSPECIFIED_SELECTION
        return reason

    @staticmethod
    def _get_caret_navigation_reason(mgr: InputEventManager) -> TextEventReason:
        """Returns the caret-navigation-granularity reason based on the last input event."""

        if mgr.last_event_was_line_navigation():
            reason = TextEventReason.NAVIGATION_BY_LINE
        elif mgr.last_event_was_word_navigation():
            reason = TextEventReason.NAVIGATION_BY_WORD
        elif mgr.last_event_was_character_navigation():
            reason = TextEventReason.NAVIGATION_BY_CHARACTER
        elif mgr.last_event_was_page_navigation():
            reason = TextEventReason.NAVIGATION_BY_PAGE
        elif mgr.last_event_was_line_boundary_navigation():
            reason = TextEventReason.NAVIGATION_TO_LINE_BOUNDARY
        elif mgr.last_event_was_file_boundary_navigation():
            reason = TextEventReason.NAVIGATION_TO_FILE_BOUNDARY
        else:
            reason = TextEventReason.UNSPECIFIED_NAVIGATION
        return reason

    @staticmethod
    def _get_editing_reason(mgr: InputEventManager) -> TextEventReason:
        """Returns the editing-operation reason based on the last input event."""

        if mgr.last_event_was_backspace():
            reason = TextEventReason.BACKSPACE
        elif mgr.last_event_was_delete():
            reason = TextEventReason.DELETE
        elif mgr.last_event_was_cut():
            reason = TextEventReason.CUT
        elif mgr.last_event_was_paste():
            reason = TextEventReason.PASTE
        elif mgr.last_event_was_undo():
            reason = TextEventReason.UNDO
        elif mgr.last_event_was_redo():
            reason = TextEventReason.REDO
        else:
            reason = TextEventReason.UNKNOWN
        return reason

    @staticmethod
    def _get_caret_moved_event_reason(event: Atspi.Event) -> TextEventReason:
        """Returns the TextEventReason for the given event."""

        from . import input_event_manager  # pylint: disable=import-outside-toplevel

        mgr = input_event_manager.get_manager()
        obj = event.source
        last_caret_set = AXUtilitiesText.get_last_caret_set()
        # Some toolkits report the resulting event's offset as -1 rather than what we set.
        if (
            last_caret_set is not None
            and last_caret_set.reason == CaretSetReason.BRAILLE_PANNING
            and obj == last_caret_set.obj
            and event.detail1 in (last_caret_set.offset, -1)
            and time.monotonic() - last_caret_set.time
            < AXUtilitiesEvent.CARET_SET_EVENT_WINDOW_SECONDS
        ):
            return TextEventReason.BRAILLE_PANNING

        mode, focus = focus_manager.get_manager().get_active_mode_and_object_of_interest()
        if mode == focus_manager.SAY_ALL:
            return TextEventReason.SAY_ALL
        if focus != obj and AXUtilitiesRole.is_text_input_search(focus):
            if mgr.last_event_was_backspace() or mgr.last_event_was_delete():
                return TextEventReason.SEARCH_UNPRESENTABLE
            return TextEventReason.SEARCH_PRESENTABLE
        is_terminal = AXUtilitiesRole.is_terminal(obj)
        if is_terminal:
            line, _start, _end = AXText.get_line_at_offset(obj)
            if AXUtilitiesEvent._is_terminal_escape_sequence(line):
                return TextEventReason.AUTO_INSERTION_UNPRESENTABLE
        if (
            mgr.last_event_was_up_or_down() or mgr.last_event_was_page_up_or_page_down()
        ) and AXUtilitiesEvent._is_spin_button_descendant(obj):
            return TextEventReason.SPIN_BUTTON_VALUE_CHANGE
        if mgr.last_event_was_caret_selection():
            return AXUtilitiesEvent._get_selection_navigation_reason(mgr)
        if mgr.last_event_was_caret_navigation():
            result = AXUtilitiesEvent._get_caret_navigation_reason(mgr)
            # For performance purposes, the input event manager does very little sanity checking
            # when determining whether an input event was line navigation. A terminal presents new
            # content via the insertion event, so a caret move that is a side effect of its redraw
            # should not be treated as navigation: paging (Less and Vim park the cursor on the
            # status line, whose transient content would otherwise be read), or auto-inserted text
            # at the prompt that looks like line navigation (e.g. pressing Up at the prompt).
            if is_terminal and (
                result == TextEventReason.NAVIGATION_BY_PAGE
                or (
                    result == TextEventReason.NAVIGATION_BY_LINE
                    and not AXUtilitiesEvent._did_line_change(obj)
                )
            ):
                result = TextEventReason.AUTO_INSERTION_UNPRESENTABLE
            return result
        if mgr.last_event_was_select_all():
            return TextEventReason.SELECT_ALL
        if mgr.last_event_was_primary_click_or_release():
            return TextEventReason.MOUSE_PRIMARY_BUTTON
        if AXUtilitiesState.is_editable(obj) or is_terminal:
            reason = AXUtilitiesEvent._get_editing_reason(mgr)
            if reason != TextEventReason.UNKNOWN:
                return reason
            if mgr.last_event_was_page_switch():
                return TextEventReason.PAGE_SWITCH
            if mgr.last_event_was_command() or mgr.last_event_was_escape():
                if focus != obj and AXUtilitiesState.is_focused(obj):
                    return TextEventReason.FOCUS_CHANGE
                return TextEventReason.UNSPECIFIED_COMMAND
            if mgr.last_event_was_printable_key():
                return TextEventReason.TYPING
            if AXUtilitiesEvent._is_terminal_being_flat_reviewed(obj):
                return TextEventReason.AUTO_INSERTION_UNPRESENTABLE
            return TextEventReason.UNKNOWN
        if mgr.last_event_was_tab_navigation() and AXUtilitiesState.is_focused(obj):
            return TextEventReason.FOCUS_CHANGE
        if AXUtilitiesObject.find_ancestor(obj, AXUtilitiesRole.children_are_presentational):
            return TextEventReason.UI_UPDATE
        return TextEventReason.UNKNOWN

    @staticmethod
    def _get_text_deletion_event_reason(event: Atspi.Event) -> TextEventReason:
        """Returns the TextEventReason for the given event."""

        from . import input_event_manager  # pylint: disable=import-outside-toplevel

        mgr = input_event_manager.get_manager()
        reason = AXUtilitiesEvent._get_obj_type_reason_or_none(event, mgr)
        if reason is not None:
            return reason

        if AXUtilitiesEvent._is_children_change(event):
            return TextEventReason.CHILDREN_CHANGE

        obj = event.source
        if not (AXUtilitiesState.is_editable(obj) or AXUtilitiesRole.is_terminal(obj)):
            return AXUtilitiesEvent._get_non_editable_reason(mgr)

        reason = AXUtilitiesEvent._get_editing_reason(mgr)
        if reason == TextEventReason.BACKSPACE:
            selected_text, _start, _end = AXUtilitiesText.get_cached_selected_text(obj)
            if event.any_data != selected_text and len(event.any_data.splitlines()) > 1:
                msg = (
                    "ERROR: Cannot speak what BackSpace deleted: any_data should contain "
                    "only the deleted text."
                )
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return TextEventReason.AUTO_DELETION_UNPRESENTABLE
        if reason != TextEventReason.UNKNOWN:
            return reason
        # Neither caret navigation nor Ctrl+Tab deletes lines. An editor which exposes only
        # part of its document is replacing what it exposes.
        if (
            not AXUtilitiesRole.is_terminal(obj)
            and "\n" in event.any_data
            and (mgr.last_event_was_caret_navigation() or mgr.last_event_was_ctrl_tab())
        ):
            return TextEventReason.AUTO_DELETION_UNPRESENTABLE
        if (
            not AXUtilitiesRole.is_terminal(obj)
            and "\n" in event.any_data
            and mgr.last_event_was_not_in_current_object(within=1.0)
        ):
            return TextEventReason.AUTO_DELETION_UNPRESENTABLE
        if mgr.last_event_was_command():
            return TextEventReason.UNSPECIFIED_COMMAND
        if mgr.last_event_was_printable_key():
            return TextEventReason.TYPING
        if mgr.last_event_was_up_or_down() or mgr.last_event_was_page_up_or_page_down():
            if AXUtilitiesEvent._is_spin_button_descendant(obj):
                return TextEventReason.SPIN_BUTTON_VALUE_CHANGE
            if (
                AXUtilitiesRole.is_terminal(obj)
                and mgr.last_event_was_up_or_down()
                and mgr.last_event_was_line_navigation()
                and len(event.any_data.split("\n")) >= 4
            ):
                AXUtilitiesEvent._save_terminal_line_navigation_deletion(event)
            return TextEventReason.AUTO_DELETION_UNPRESENTABLE

        selected_text, _start, _end = AXUtilitiesText.get_cached_selected_text(obj)
        if selected_text and event.any_data.strip() == selected_text.strip():
            return TextEventReason.SELECTED_TEXT_DELETION
        if AXUtilitiesEvent._is_terminal_being_flat_reviewed(obj):
            return TextEventReason.AUTO_DELETION_UNPRESENTABLE
        return TextEventReason.UNKNOWN

    @staticmethod
    def _get_text_insertion_event_reason(event: Atspi.Event) -> TextEventReason:
        """Returns the TextEventReason for the given event."""

        from . import (
            input_event_manager,  # pylint: disable=import-outside-toplevel
            typing_echo_presenter,  # pylint: disable=import-outside-toplevel
        )

        mgr = input_event_manager.get_manager()
        reason = AXUtilitiesEvent._get_obj_type_reason_or_none(event, mgr)
        if reason is not None:
            return reason

        if AXUtilitiesEvent._is_children_change(event):
            return TextEventReason.CHILDREN_CHANGE

        obj = event.source
        is_terminal = AXUtilitiesRole.is_terminal(obj)
        if not (AXUtilitiesState.is_editable(obj) or is_terminal):
            return AXUtilitiesEvent._get_non_editable_reason(mgr)

        if is_terminal and AXUtilitiesEvent._is_terminal_escape_sequence(event.any_data):
            return TextEventReason.AUTO_INSERTION_UNPRESENTABLE

        selected_text, _start, _end = AXUtilitiesText.get_selected_text(obj)
        has_selected = bool(selected_text and event.any_data == selected_text)

        if mgr.last_event_was_backspace():
            return TextEventReason.BACKSPACE
        if mgr.last_event_was_delete():
            return TextEventReason.DELETE
        if mgr.last_event_was_cut():
            return TextEventReason.CUT
        if mgr.last_event_was_paste():
            return TextEventReason.PASTE
        if mgr.last_event_was_undo():
            return (
                TextEventReason.SELECTED_TEXT_RESTORATION if has_selected else TextEventReason.UNDO
            )
        if mgr.last_event_was_redo():
            return (
                TextEventReason.SELECTED_TEXT_RESTORATION if has_selected else TextEventReason.REDO
            )
        # Caret navigation does not insert text. An editor which exposes only part of its
        # document is replacing what it exposes. We present the new location in response to
        # the caret-moved event. Single-line objects are excluded because arrowing through
        # an autocompletion list does insert text which should be presented. Spin buttons
        # are excluded because Home, End, and Page Up/Down change their value.
        is_document = (
            not is_terminal
            and not AXUtilitiesState.is_single_line(obj)
            and not AXUtilitiesEvent._is_spin_button_descendant(obj)
        )
        if is_document and (
            mgr.last_event_was_caret_navigation()
            or mgr.last_event_was_not_in_current_object(within=1.0)
        ):
            return TextEventReason.AUTO_INSERTION_UNPRESENTABLE
        if not is_terminal and "\n" in event.any_data and mgr.last_event_was_ctrl_tab():
            return TextEventReason.AUTO_INSERTION_UNPRESENTABLE
        if mgr.last_event_was_command():
            return TextEventReason.UNSPECIFIED_COMMAND
        if mgr.last_event_was_space() and not AXUtilitiesRole.is_password_text(obj):
            if event.any_data == "\n":
                return TextEventReason.AUTO_INSERTION_UNPRESENTABLE
            return TextEventReason.TYPING
        if mgr.last_event_was_tab() or mgr.last_event_was_return():
            if not event.any_data.strip():
                return TextEventReason.TYPING
            if mgr.last_event_was_tab():
                return TextEventReason.AUTO_INSERTION_PRESENTABLE
            if AXUtilitiesState.is_single_line(obj):
                return TextEventReason.AUTO_INSERTION_UNPRESENTABLE
            # Text which Return inserts, be it a newline or an accepted completion, leaves the
            # caret at the end of that text. A terminal is excluded: there Return runs a command
            # rather than accepting a completion, so this caret check does not apply.
            if not is_terminal and AXText.get_caret_offset(obj) != event.detail1 + event.detail2:
                return TextEventReason.AUTO_INSERTION_UNPRESENTABLE
            return TextEventReason.AUTO_INSERTION_PRESENTABLE
        if mgr.last_event_was_printable_key() or mgr.last_event_was_space():
            if has_selected:
                return TextEventReason.AUTO_INSERTION_PRESENTABLE
            presenter = typing_echo_presenter.get_presenter()
            if AXUtilitiesRole.is_password_text(obj):
                echo = presenter.get_key_echo_enabled()
            else:
                echo = presenter.get_character_echo_enabled()
            return TextEventReason.TYPING_ECHOABLE if echo else TextEventReason.TYPING
        if mgr.last_event_was_middle_click() or mgr.last_event_was_middle_release():
            return TextEventReason.MOUSE_MIDDLE_BUTTON
        if mgr.last_event_was_up_or_down() or mgr.last_event_was_page_up_or_page_down():
            if AXUtilitiesEvent._is_spin_button_descendant(obj):
                return TextEventReason.SPIN_BUTTON_VALUE_CHANGE
            if (
                is_terminal
                and mgr.last_event_was_up_or_down()
                and mgr.last_event_was_line_navigation()
            ):
                if AXUtilitiesEvent._is_terminal_line_navigation_repaint(event):
                    AXUtilitiesEvent._save_terminal_line_navigation_repaint_line(event)
                    return TextEventReason.TERMINAL_LINE_NAVIGATION_REPAINT
                # VTE can draw the row scrolled into view as a separate insertion after the
                # repaint. If the repaint handling already presented exactly that line, saying
                # it again is noise.
                if AXUtilitiesEvent._repeats_terminal_line_navigation_repaint_line(event):
                    return TextEventReason.AUTO_INSERTION_UNPRESENTABLE
            return TextEventReason.AUTO_INSERTION_PRESENTABLE
        if has_selected:
            return TextEventReason.SELECTED_TEXT_INSERTION
        if AXUtilitiesEvent._is_terminal_being_flat_reviewed(obj):
            return TextEventReason.AUTO_INSERTION_UNPRESENTABLE
        return TextEventReason.UNKNOWN

    @staticmethod
    def _get_text_selection_changed_event_reason(event: Atspi.Event) -> TextEventReason:
        """Returns the TextEventReason for the given event."""

        from . import input_event_manager  # pylint: disable=import-outside-toplevel

        mgr = input_event_manager.get_manager()
        obj = event.source
        focus = focus_manager.get_manager().get_locus_of_focus()

        if focus != obj and AXUtilitiesRole.is_text_input_search(focus):
            if mgr.last_event_was_backspace() or mgr.last_event_was_delete():
                return TextEventReason.SEARCH_UNPRESENTABLE
            return TextEventReason.SEARCH_PRESENTABLE
        if mgr.last_event_was_caret_selection():
            return AXUtilitiesEvent._get_selection_navigation_reason(mgr)
        if mgr.last_event_was_caret_navigation():
            return AXUtilitiesEvent._get_caret_navigation_reason(mgr)
        if mgr.last_event_was_select_all():
            return TextEventReason.SELECT_ALL
        if mgr.last_event_was_primary_click_or_release():
            return TextEventReason.MOUSE_PRIMARY_BUTTON
        if not (AXUtilitiesState.is_editable(obj) or AXUtilitiesRole.is_terminal(obj)):
            return TextEventReason.UNKNOWN

        reason = AXUtilitiesEvent._get_editing_reason(mgr)
        if reason != TextEventReason.UNKNOWN:
            return reason
        if mgr.last_event_was_page_switch():
            return TextEventReason.PAGE_SWITCH
        if mgr.last_event_was_command():
            return TextEventReason.UNSPECIFIED_COMMAND
        if mgr.last_event_was_printable_key():
            cached_text, old_start, old_end = AXUtilitiesText.get_cached_selected_text(obj)
            if not cached_text:
                _current, new_start, new_end = AXUtilitiesText.get_selected_text(obj)
                if new_start < new_end:
                    return TextEventReason.AUTO_SELECTION
            else:
                _current, new_start, new_end = AXUtilitiesText.get_selected_text(obj)
                if new_start == old_start + 1 and new_end == old_end:
                    return TextEventReason.AUTO_UNSELECTION
            return TextEventReason.TYPING
        if mgr.last_event_was_up_or_down() or mgr.last_event_was_page_up_or_page_down():
            if AXUtilitiesEvent._is_spin_button_descendant(obj):
                return TextEventReason.SPIN_BUTTON_VALUE_CHANGE
        return TextEventReason.UNKNOWN

    @staticmethod
    def is_presentable_active_descendant_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as an active-descendant change."""

        if not event.any_data:
            msg = "AXUtilitiesEvent: No any_data."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if not (
            AXUtilitiesState.is_focused(event.source) or AXUtilitiesState.is_focused(event.any_data)
        ):
            msg = "AXUtilitiesEvent: Neither source nor child have focused state."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        focus = focus_manager.get_manager().get_locus_of_focus()
        if AXUtilitiesRole.is_table_cell(focus):
            table = AXUtilitiesObject.find_ancestor(focus, AXUtilitiesRole.is_tree_or_tree_table)
            if table is not None and table != event.source:
                msg = "AXUtilitiesEvent: Event is from a different tree or tree table."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_checked_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as a checked-state change."""

        old_state = AXUtilitiesEvent._CACHE.get_state(
            AXUtilitiesEvent._CACHE.LAST_KNOWN_CHECKED,
            event.source,
        )
        new_state = AXUtilitiesState.is_checked(event.source)
        if old_state == new_state:
            msg = "AXUtilitiesEvent: The new state matches the old state."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        AXUtilitiesEvent._CACHE.set_state(
            AXUtilitiesEvent._CACHE.LAST_KNOWN_CHECKED,
            event.source,
            new_state,
        )
        focus = focus_manager.get_manager().get_locus_of_focus()
        if event.source != focus:
            if not AXUtilitiesObject.is_ancestor(event.source, focus):
                msg = "AXUtilitiesEvent: The source is not the locus of focus or its descendant."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False
            if not (AXUtilitiesRole.is_list_item(focus) or AXUtilitiesRole.is_tree_item(focus)):
                msg = "AXUtilitiesEvent: The source descends from non-interactive-item focus."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False

        from . import input_event_manager  # pylint: disable=import-outside-toplevel

        mgr = input_event_manager.get_manager()

        # Radio buttons normally change their state when you arrow to them, so we handle the
        # announcement of their state changes in the focus handling code.
        if AXUtilitiesRole.is_radio_button(event.source) and not mgr.last_event_was_space():
            msg = "AXUtilitiesEvent: Only presentable for this role if toggled by user."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def _is_changed_description_to_present(event: Atspi.Event) -> bool:
        """Returns True if the description data warrants further presentability checks."""

        if not isinstance(event.any_data, str):
            msg = "AXUtilitiesEvent: The any_data is not a string."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        old_description = AXUtilitiesEvent._CACHE.get_description(event.source)
        new_description = event.any_data
        if old_description == new_description:
            msg = "AXUtilitiesEvent: The new description matches the old description."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if AXUtilitiesEvent._strings_are_redundant(old_description, new_description):
            msg = (
                f"AXUtilitiesEvent: The new description ('{new_description}') "
                f"is too similar to the old description ('{old_description}')."
            )
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        AXUtilitiesEvent._CACHE.set_description(event.source, new_description)
        if not new_description:
            msg = "AXUtilitiesEvent: The description is empty."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        return True

    @staticmethod
    def is_presentable_description_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as a description change."""

        if not AXUtilitiesEvent._is_changed_description_to_present(event):
            return False

        if not AXUtilitiesState.is_showing(event.source):
            msg = "AXUtilitiesEvent: The event source is not showing."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        manager = focus_manager.get_manager()
        focus = manager.get_locus_of_focus()
        if event.source != focus and not AXUtilitiesObject.is_ancestor(focus, event.source):
            msg = "AXUtilitiesEvent: The event is not from the locus of focus or ancestor."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        window = manager.get_active_window()
        if event.source != window and event.any_data == AXObject.get_name(window):
            msg = "AXUtilitiesEvent: The new description matches the active window's name."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_expanded_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as an expanded-state change."""

        old_state = AXUtilitiesEvent._CACHE.get_state(
            AXUtilitiesEvent._CACHE.LAST_KNOWN_EXPANDED,
            event.source,
        )
        new_state = AXUtilitiesState.is_expanded(event.source)
        if old_state == new_state:
            msg = "AXUtilitiesEvent: The new state matches the old state."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        AXUtilitiesEvent._CACHE.set_state(
            AXUtilitiesEvent._CACHE.LAST_KNOWN_EXPANDED,
            event.source,
            new_state,
        )
        focus = focus_manager.get_manager().get_locus_of_focus()
        if event.source == focus:
            msg = "AXUtilitiesEvent: Event is presentable, from the locus of focus."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if not event.detail1 and not AXUtilitiesObject.is_ancestor(focus, event.source):
            msg = "AXUtilitiesEvent: Event is not from the locus of focus or ancestor."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if AXUtilitiesRole.is_table_row(event.source) or AXUtilitiesRole.is_list_box(event.source):
            msg = "AXUtilitiesEvent: Event is presentable based on role."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if AXUtilitiesRole.is_combo_box(event.source) or AXUtilitiesRole.is_button(event.source):
            if not AXUtilitiesState.is_focused(event.source):
                msg = "AXUtilitiesEvent: Only presentable for this role if focused."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_indeterminate_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as an indeterminate-state change."""

        old_state = AXUtilitiesEvent._CACHE.get_state(
            AXUtilitiesEvent._CACHE.LAST_KNOWN_INDETERMINATE,
            event.source,
        )
        new_state = AXUtilitiesState.is_indeterminate(event.source)
        if old_state == new_state:
            msg = "AXUtilitiesEvent: The new state matches the old state."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        AXUtilitiesEvent._CACHE.set_state(
            AXUtilitiesEvent._CACHE.LAST_KNOWN_INDETERMINATE,
            event.source,
            new_state,
        )

        # If this state is cleared, the new state will become checked or unchecked
        # and we should get object:state-changed:checked events for those cases.
        if not new_state:
            msg = "AXUtilitiesEvent: The new state should be presented as a checked-change."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if event.source != focus_manager.get_manager().get_locus_of_focus():
            msg = "AXUtilitiesEvent: The event is not from the locus of focus."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_invalid_entry_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as an invalid-entry-state change."""

        old_state = AXUtilitiesEvent._CACHE.get_state(
            AXUtilitiesEvent._CACHE.LAST_KNOWN_INVALID_ENTRY,
            event.source,
        )
        new_state = AXUtilitiesState.is_invalid_entry(event.source)
        if old_state == new_state:
            msg = "AXUtilitiesEvent: The new state matches the old state."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        AXUtilitiesEvent._CACHE.set_state(
            AXUtilitiesEvent._CACHE.LAST_KNOWN_INVALID_ENTRY,
            event.source,
            new_state,
        )
        if event.source != focus_manager.get_manager().get_locus_of_focus():
            msg = "AXUtilitiesEvent: The event is not from the locus of focus."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def _is_presentable_frame_name_change(event: Atspi.Event) -> bool:
        """Returns True if a frame's name-change event should be presented."""

        if event.source != focus_manager.get_manager().get_active_window():
            msg = "AXUtilitiesEvent: Event is for frame other than the active window."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        # Example: Typing the subject in an email client causing the window name to change.
        focus = focus_manager.get_manager().get_locus_of_focus()
        if (
            AXUtilitiesState.is_editable(focus)
            and AXText.get_character_count(focus)
            and AXText.get_all_text(focus) in event.any_data
        ):
            msg = "AXUtilitiesEvent: Event is redundant notification for the locus of focus."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if AXUtilitiesRole.is_terminal(focus):
            text = AXText.get_line_at_offset(focus)[0].strip()
            if AXUtilitiesEvent._strings_are_redundant(text, event.any_data):
                msg = (
                    f"AXUtilitiesEvent: The new name ('{event.any_data}') "
                    f"is too similar to the text at offset ('{text}')."
                )
                debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def _is_changed_name_to_present(event: Atspi.Event) -> bool:
        """Returns True if the name data warrants further presentability checks."""

        if AXUtilitiesEvent._CACHE.is_ignoring_name_changes_for(event.source):
            msg = "AXUtilitiesEvent: Ignoring name change for this source."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if not isinstance(event.any_data, str):
            msg = "AXUtilitiesEvent: The any_data is not a string."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        old_name = AXUtilitiesEvent._CACHE.get_name(event.source)
        new_name = event.any_data
        if old_name == new_name:
            msg = "AXUtilitiesEvent: The new name matches the old name."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if AXUtilitiesEvent._strings_are_redundant(old_name, new_name):
            msg = (
                f"AXUtilitiesEvent: The new name ('{new_name}') "
                f"is too similar to the old name ('{old_name}')."
            )
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        AXUtilitiesEvent._CACHE.set_name(event.source, new_name)
        if not new_name:
            msg = "AXUtilitiesEvent: The name is empty."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        return True

    @staticmethod
    def is_presentable_name_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as a name change."""

        if not AXUtilitiesEvent._is_changed_name_to_present(event):
            return False

        if not AXUtilitiesState.is_showing(event.source):
            msg = "AXUtilitiesEvent: The event source is not showing."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            if event.source != focus_manager.get_manager().get_locus_of_focus():
                return False

        if AXUtilitiesRole.is_frame(event.source):
            return AXUtilitiesEvent._is_presentable_frame_name_change(event)

        if event.source != focus_manager.get_manager().get_locus_of_focus():
            msg = "AXUtilitiesEvent: The event is not from the locus of focus."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if AXUtilitiesRole.is_list_item(event.source):
            # Inspired by Firefox's downloads list in which the list item's name changes to show
            # download progress. It's super chatty and stomps on the progress bar announcements.
            if AXObject.supports_collection(event.source):
                has_progress_bar = (
                    AXUtilitiesCollection.find_first_with_role(
                        event.source, [Atspi.Role.PROGRESS_BAR]
                    )
                    is not None
                )
            else:
                has_progress_bar = (
                    AXUtilitiesObject.find_descendant(event.source, AXUtilitiesRole.is_progress_bar)
                    is not None
                )
            if has_progress_bar:
                msg = "AXUtilitiesEvent: The list item contains a progress bar."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                AXUtilitiesEvent._CACHE.ignore_name_changes_for(event.source)
                return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_pressed_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as a pressed-state change."""

        old_state = AXUtilitiesEvent._CACHE.get_state(
            AXUtilitiesEvent._CACHE.LAST_KNOWN_PRESSED,
            event.source,
        )
        new_state = AXUtilitiesState.is_pressed(event.source)
        if old_state == new_state:
            msg = "AXUtilitiesEvent: The new state matches the old state."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        AXUtilitiesEvent._CACHE.set_state(
            AXUtilitiesEvent._CACHE.LAST_KNOWN_PRESSED,
            event.source,
            new_state,
        )
        if event.source != focus_manager.get_manager().get_locus_of_focus():
            msg = "AXUtilitiesEvent: The event is not from the locus of focus."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_selected_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as a selected-state change."""

        old_state = AXUtilitiesEvent._CACHE.get_state(
            AXUtilitiesEvent._CACHE.LAST_KNOWN_SELECTED,
            event.source,
        )
        new_state = AXUtilitiesState.is_selected(event.source)
        if old_state == new_state:
            msg = "AXUtilitiesEvent: The new state matches the old state."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        AXUtilitiesEvent._CACHE.set_state(
            AXUtilitiesEvent._CACHE.LAST_KNOWN_SELECTED,
            event.source,
            new_state,
        )
        if event.source != focus_manager.get_manager().get_locus_of_focus():
            msg = "AXUtilitiesEvent: The event is not from the locus of focus."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_selection_change(event: Atspi.Event) -> bool:
        """Returns True if this object:selection-changed event should be presented."""

        from . import input_event_manager  # pylint: disable=import-outside-toplevel

        if input_event_manager.get_manager().last_event_was_space():
            msg = "AXUtilitiesEvent: Selection toggled via space; handled elsewhere."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        source_states = AXObject.get_state_set(event.source)
        if AXUtilitiesState.manages_descendants(event.source, source_states):
            msg = "AXUtilitiesEvent: Source manages descendants; handled elsewhere."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if AXUtilitiesRole.is_menu(event.source):
            if AXUtilitiesState.is_showing_and_visible(event.source, source_states):
                msg = "AXUtilitiesEvent: Event is presentable: Source is a menu."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True

            # This is a sad workaround for GTK2 menu items with submenus losing their showing state
            # when submenu children become selected.
            child = AXSelection.get_selected_child(event.source, 0)
            if child is not None and AXUtilitiesState.is_showing_and_visible(child):
                tokens = [
                    "AXUtilitiesEvent: Event is presentable: Selected child",
                    child,
                    "is showing and visible",
                ]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return True

            tokens = ["AXUtilitiesEvent: Menu lacks showing + visible:", event.source]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        if AXUtilitiesRole.is_combo_box(event.source) and not AXUtilitiesState.is_expanded(
            event.source, source_states
        ):
            text_input = AXUtilitiesObject.find_descendant(
                event.source, AXUtilitiesRole.is_text_input
            )
            if text_input is not None and AXUtilitiesState.is_focused(text_input):
                msg = "AXUtilitiesEvent: Combo box text input is focused; not a user selection."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False

        focus = focus_manager.get_manager().get_locus_of_focus()

        mgr = input_event_manager.get_manager()
        if (
            AXUtilitiesState.supports_autocompletion(focus)
            and not mgr.last_event_was_up_or_down()
            and not mgr.last_event_was_page_up_or_page_down()
        ):
            if AXUtilitiesRelation.object_is_controlled_by(event.source, focus):
                msg = "AXUtilitiesEvent: Source is autocomplete for focused widget; not presenting."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False
        if event.source != focus and not AXUtilitiesState.is_showing_and_visible(
            event.source, source_states
        ):
            combobox = AXUtilitiesObject.find_ancestor(event.source, AXUtilitiesRole.is_combo_box)
            if combobox != focus and event.source != AXObject.get_parent(focus):
                tokens = ["AXUtilitiesEvent: Source lacks showing + visible:", event.source]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return False

        active_window = focus_manager.get_manager().get_active_window()
        if active_window and not AXUtilitiesObject.is_ancestor(event.source, active_window):
            tokens = ["AXUtilitiesEvent:", event.source, "is not inside", active_window]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        msg = "AXUtilitiesEvent: Event is presentable: Not ruled out by any checks."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_value_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as a value change."""

        old_value = AXUtilitiesEvent._CACHE.get_value(event.source)
        new_value = AXValue.get_current_value(event.source)
        AXUtilitiesEvent._CACHE.set_value(event.source, new_value)

        old_value_text = AXUtilitiesEvent._CACHE.get_value_text(event.source)
        new_value_text = AXValue.get_current_value_text(event.source)
        AXUtilitiesEvent._CACHE.set_value_text(event.source, new_value_text)

        if old_value == new_value and old_value_text == new_value_text:
            msg = "AXUtilitiesEvent: The new value matches the old value."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if AXUtilitiesRole.is_progress_bar(event.source):
            if AXUtilitiesComponent.has_no_size(event.source):
                tokens = ["AXUtilitiesEvent:", event.source, "has no size."]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return False
            if AXUtilitiesObject.find_ancestor(event.source, AXUtilitiesRole.is_status_bar):
                tokens = ["AXUtilitiesEvent:", event.source, "is in a status bar."]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def _is_presentable_text_event(event: Atspi.Event) -> bool:
        """Returns True if this text event should be presented."""

        source_states = AXObject.get_state_set(event.source)
        if not (
            AXUtilitiesState.is_editable(event.source, source_states)
            or AXUtilitiesRole.is_terminal(event.source)
        ):
            msg = "AXUtilitiesEvent: The source is neither editable nor a terminal."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        focus = focus_manager.get_manager().get_locus_of_focus()
        if focus != event.source and not AXUtilitiesState.is_focused(event.source, source_states):
            msg = "AXUtilitiesEvent: The source is neither focused, nor the locus of focus"
            debug.print_message(debug.LEVEL_INFO, msg, True)

            # This can happen in web content where the focus is a contenteditable element and a
            # new child element is created for new or changed text.
            if AXUtilitiesObject.is_ancestor(event.source, focus):
                msg = "AXUtilitiesEvent: The locus of focus is an ancestor of the source."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True

            return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_text_attributes_change(event: Atspi.Event) -> bool:
        """Returns True if this text-attributes-change event should be presented."""

        return AXUtilitiesEvent._is_presentable_text_event(event)

    @staticmethod
    def is_presentable_text_deletion(event: Atspi.Event) -> bool:
        """Returns True if this text-deletion event should be presented."""

        return AXUtilitiesEvent._is_presentable_text_event(event)

    @staticmethod
    def is_presentable_text_insertion(event: Atspi.Event) -> bool:
        """Returns True if this text-insertion event should be presented."""

        return AXUtilitiesEvent._is_presentable_text_event(event)



# Orca
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Preferences grid for MathCAT presentation support."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import guilabels, math_presenter, preferences_grid_base

if TYPE_CHECKING:
    from .math_presenter import MathPresenter


class MathPreferencesGrid(preferences_grid_base.AutoPreferencesGrid):
    """Sub-grid for math settings within the Documents page."""

    _gsettings_schema = "math-presentation"
    _documentation_summary = (
        "Use these settings to control speech, braille, copying, and navigation for math content."
    )

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for math presentation preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.MATH_PRESENTATION,
            panel_id="math_presenter.math-presentation",
            description=(
                "Math presentation settings control how Orca speaks, brailles, copies, "
                "and navigates mathematical expressions."
            ),
            schema="math-presentation",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.LANGUAGE,
                    kind="combo",
                    summary="Selects the language Orca uses for math presentation.",
                    schema="math-presentation",
                    key=math_presenter.MathPresenter.KEY_LANGUAGE,
                    dynamic_values=True,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.MATH_SPEECH_STYLE,
                    kind="combo",
                    summary="Selects the style Orca uses when speaking math.",
                    schema="math-presentation",
                    key=math_presenter.MathPresenter.KEY_SPEECH_STYLE,
                    dynamic_values=True,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.VERBOSITY,
                    kind="combo",
                    summary="Selects how much detail Orca speaks for math.",
                    schema="math-presentation",
                    key=math_presenter.MathPresenter.KEY_VERBOSITY,
                    value_docs=(
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.MATH_VERBOSITY_TERSE,
                            value="Terse",
                            summary="Speaks less detail.",
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.MATH_VERBOSITY_MEDIUM,
                            value="Medium",
                            summary="Speaks a moderate amount of detail.",
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.MATH_VERBOSITY_VERBOSE,
                            value="Verbose",
                            summary="Speaks the most detail.",
                        ),
                    ),
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.MATH_BRAILLE_CODE,
                    kind="combo",
                    summary=("Selects the braille code Orca uses for mathematical expressions."),
                    schema="math-presentation",
                    key=math_presenter.MathPresenter.KEY_BRAILLE_CODE,
                    dynamic_values=True,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.MATH_COPY_FORMAT,
                    kind="combo",
                    summary="Selects the format used when copying math content.",
                    schema="math-presentation",
                    key=math_presenter.MathPresenter.KEY_COPY_FORMAT,
                    values=("MathML", "LaTeX", "ASCIIMath", guilabels.SPEECH),
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.MATH_NAV_MODE,
                    kind="combo",
                    summary="Selects how Orca moves through math during navigation.",
                    schema="math-presentation",
                    key=math_presenter.MathPresenter.KEY_NAV_MODE,
                    value_docs=(
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.MATH_NAV_MODE_ENHANCED,
                            value="enhanced",
                            summary="Groups related items together for faster navigation.",
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.MATH_NAV_MODE_SIMPLE,
                            value="simple",
                            summary="Visits each math element individually.",
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.MATH_NAV_MODE_CHARACTER,
                            value="character",
                            summary="Visits individual symbols one at a time.",
                        ),
                    ),
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.MATH_BRAILLE_NAV_HIGHLIGHT,
                    kind="combo",
                    summary=("Selects how Orca marks the current math position in braille."),
                    schema="math-presentation",
                    key=math_presenter.MathPresenter.KEY_BRAILLE_NAV_HIGHLIGHT,
                    value_docs=(
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.MATH_BRAILLE_HIGHLIGHT_NONE,
                            value="off",
                            summary="Does not mark the current math position.",
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.MATH_BRAILLE_HIGHLIGHT_FIRST_CHAR,
                            value="first-char",
                            summary="Marks the first cell of the current math item.",
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.MATH_BRAILLE_HIGHLIGHT_END_POINTS,
                            value="end-points",
                            summary="Marks the first and last cells of the current math item.",
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.MATH_BRAILLE_HIGHLIGHT_ALL,
                            value="all",
                            summary="Marks every cell of the current math item.",
                        ),
                    ),
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.MATH_AUTO_ZOOM_OUT,
                    kind="switch",
                    summary=(
                        "Controls whether Orca automatically exits a two-dimensional math "
                        "structure when you navigate past its edge."
                    ),
                    schema="math-presentation",
                    key=math_presenter.MathPresenter.KEY_AUTO_ZOOM_OUT,
                ),
            ),
        )

    def __init__(self, presenter: MathPresenter) -> None:
        languages = presenter.get_language_choices()
        speech_styles = presenter.get_speech_style_choices()
        braille_codes = presenter.get_braille_code_choices()
        verbosity_labels = [
            guilabels.MATH_VERBOSITY_TERSE,
            guilabels.MATH_VERBOSITY_MEDIUM,
            guilabels.MATH_VERBOSITY_VERBOSE,
        ]
        verbosity_values = ["Terse", "Medium", "Verbose"]
        controls: list[preferences_grid_base.ControlType] = [
            preferences_grid_base.EnumPreferenceControl(
                label=guilabels.LANGUAGE,
                options=languages,
                values=languages,
                getter=presenter.get_language,
                setter=presenter.set_language,
                prefs_key=math_presenter.MathPresenter.KEY_LANGUAGE,
            ),
            preferences_grid_base.EnumPreferenceControl(
                label=guilabels.MATH_SPEECH_STYLE,
                options=speech_styles,
                values=speech_styles,
                getter=presenter.get_speech_style,
                setter=presenter.set_speech_style,
                prefs_key=math_presenter.MathPresenter.KEY_SPEECH_STYLE,
            ),
            preferences_grid_base.EnumPreferenceControl(
                label=guilabels.VERBOSITY,
                options=verbosity_labels,
                values=verbosity_values,
                getter=presenter.get_verbosity,
                setter=presenter.set_verbosity,
                prefs_key=math_presenter.MathPresenter.KEY_VERBOSITY,
            ),
            preferences_grid_base.EnumPreferenceControl(
                label=guilabels.MATH_BRAILLE_CODE,
                options=braille_codes,
                values=braille_codes,
                getter=presenter.get_braille_code,
                setter=presenter.set_braille_code,
                prefs_key=math_presenter.MathPresenter.KEY_BRAILLE_CODE,
            ),
            preferences_grid_base.EnumPreferenceControl(
                label=guilabels.MATH_COPY_FORMAT,
                options=["MathML", "LaTeX", "ASCIIMath", guilabels.SPEECH],
                values=["mathml", "latex", "asciimath", "speech"],
                getter=presenter.get_copy_format,
                setter=presenter.set_copy_format,
                prefs_key=math_presenter.MathPresenter.KEY_COPY_FORMAT,
            ),
            preferences_grid_base.EnumPreferenceControl(
                label=guilabels.MATH_NAV_MODE,
                options=[
                    guilabels.MATH_NAV_MODE_ENHANCED,
                    guilabels.MATH_NAV_MODE_SIMPLE,
                    guilabels.MATH_NAV_MODE_CHARACTER,
                ],
                values=["enhanced", "simple", "character"],
                getter=presenter.get_nav_mode,
                setter=presenter.set_nav_mode,
                prefs_key=math_presenter.MathPresenter.KEY_NAV_MODE,
            ),
            preferences_grid_base.EnumPreferenceControl(
                label=guilabels.MATH_BRAILLE_NAV_HIGHLIGHT,
                options=[
                    guilabels.MATH_BRAILLE_HIGHLIGHT_NONE,
                    guilabels.MATH_BRAILLE_HIGHLIGHT_FIRST_CHAR,
                    guilabels.MATH_BRAILLE_HIGHLIGHT_END_POINTS,
                    guilabels.MATH_BRAILLE_HIGHLIGHT_ALL,
                ],
                values=["off", "first-char", "end-points", "all"],
                getter=presenter.get_braille_nav_highlight,
                setter=presenter.set_braille_nav_highlight,
                prefs_key=math_presenter.MathPresenter.KEY_BRAILLE_NAV_HIGHLIGHT,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.MATH_AUTO_ZOOM_OUT,
                getter=presenter.get_auto_zoom_out,
                setter=presenter.set_auto_zoom_out,
                prefs_key=math_presenter.MathPresenter.KEY_AUTO_ZOOM_OUT,
            ),
        ]
        super().__init__(guilabels.MATH_PRESENTATION, controls)



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Handles writing debugging messages to the debug file or stderr."""

import contextlib
import inspect
import os
import re
import sys
import threading
import traceback
from datetime import datetime, timezone
from typing import Any, TextIO

LEVEL_OFF = 10000
LEVEL_SEVERE = 1000
LEVEL_WARNING = 900
LEVEL_INFO = 800
LEVEL_ALL = 0

# Leave these as-is for now so as not to break debugging for anyone using orca-customizations.py
# pylint: disable=invalid-name
debugLevel: int = LEVEL_SEVERE
debugFile: TextIO | None = None
# pylint: enable=invalid-name

_printing = threading.local()


def print_exception(level: int) -> None:
    """Prints out information regarding the current exception."""
    if level >= debugLevel:
        _print_text(level)
        traceback.print_exc(100, debugFile)
        _print_text(level)


def print_tokens(
    level: int,
    tokens: list[Any],
    timestamp: bool = False,
    stack: bool = False,
) -> None:
    """Prints out each token as a human-consumable string."""

    if level < debugLevel:
        return

    # Avoid importing AXObject through AXUtilitiesDebugging before AXCacheManager initializes.
    # pylint: disable=import-outside-toplevel
    from .ax_utilities_debugging import AXUtilitiesDebugging

    text = " ".join(map(AXUtilitiesDebugging.as_string, tokens))
    text = re.sub(r" (?=[,.:)])(?![\n])", "", text)
    _print_text(level, text, timestamp, stack)


def print_message(level: int, text: str, timestamp: bool = False, stack: bool = False) -> None:
    """Prints out text."""

    if level < debugLevel:
        return

    _print_text(level, text, timestamp, stack)


def shutdown() -> None:
    """Flushes debugFile and fsyncs it so no buffered writes are lost on exit."""

    if debugFile is None:
        return
    with contextlib.suppress(OSError):
        debugFile.flush()
        os.fsync(debugFile.fileno())


def _stack_as_string(max_frames: int = 4) -> str:
    # Avoid importing AXObject through AXUtilitiesDebugging before AXCacheManager initializes.
    # pylint: disable=import-outside-toplevel
    from .ax_utilities_debugging import AXUtilitiesDebugging

    callers = []
    current_module = inspect.getmodule(inspect.currentframe())
    stack = inspect.stack()
    for i in range(1, len(stack)):
        frame = stack[i]
        module = inspect.getmodule(frame[0])
        if module == current_module:
            continue
        if frame.function == "main":
            continue
        if module is None or module.__name__ is None:
            continue
        callers.append(frame)
        if len(callers) >= max_frames:
            break

    callers.reverse()
    return " > ".join(map(AXUtilitiesDebugging.as_string, callers))


def _print_text(level: int, text: str = "", timestamp: bool = False, stack: bool = False) -> None:
    if level < debugLevel:
        return

    # Prevent reentrancy.
    if getattr(_printing, "active", False):
        return

    _printing.active = True

    try:
        if timestamp:
            text = text.replace("\n", f"\n{' ' * 18}")
            local_time = datetime.now(tz=timezone.utc).astimezone()
            text = f"{local_time.strftime('%H:%M:%S.%f')} - {text}"
        if stack:
            text += f" {_stack_as_string()}"

        if debugFile:
            try:
                debugFile.write(f"{text}\n")
            except (AttributeError, OSError):
                return
            except (TypeError, ValueError, UnicodeEncodeError) as error:
                text = f"Exception trying to write text to file: {error}"
                with contextlib.suppress(
                    AttributeError, OSError, TypeError, ValueError, UnicodeEncodeError
                ):
                    debugFile.write(f"{text}\n")
            if level >= LEVEL_SEVERE:
                with contextlib.suppress(
                    AttributeError, OSError, TypeError, ValueError, UnicodeEncodeError
                ):
                    sys.stderr.write(f"{text}\n")
        else:
            try:
                sys.stderr.write(f"{text}\n")
            except (AttributeError, OSError):
                return
            except (TypeError, ValueError, UnicodeEncodeError) as error:
                text = f"Exception trying to write text to stderr: {error}"
                with contextlib.suppress(
                    AttributeError, OSError, TypeError, ValueError, UnicodeEncodeError
                ):
                    sys.stderr.write(f"{text}\n")
    finally:
        _printing.active = False



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2011-2025 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-public-methods

"""Provides sound presentation support."""

from __future__ import annotations

import time
from enum import Enum
from typing import TYPE_CHECKING, Any

from . import (
    dbus_service,
    debug,
    document_presenter,
    focus_manager,
    gsettings_registry,
    sound,
)
from .generator import GeneratorContext, PresentationReason

if TYPE_CHECKING:
    from collections.abc import Callable

    import gi

    from .dbus_service import UInt32
    from .scripts import default
    from .sound import Icon, Tone
    from .sound_presenter_preferences_grid import SoundPreferencesGrid

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi


@gsettings_registry.get_registry().gsettings_enum(
    "org.gnome.Orca.ProgressBarVerbosity",
    values={"all": 0, "application": 1, "window": 2},
)
class ProgressBarVerbosity(Enum):
    """Progress bar verbosity level enumeration."""

    ALL = 0
    APPLICATION = 1
    WINDOW = 2

    @property
    def string_name(self) -> str:
        """Returns the lowercase string name for this enum value."""

        return self.name.lower()


@gsettings_registry.get_registry().gsettings_schema("org.gnome.Orca.Sound", name="sound")
class SoundPresenter:
    """Provides sound presentation support."""

    _SCHEMA = "sound"
    KEY_ENABLED = "enabled"
    KEY_VOLUME = "volume"
    KEY_BEEP_PROGRESS_BAR_UPDATES = "beep-progress-bar-updates"
    KEY_PROGRESS_BAR_BEEP_INTERVAL = "progress-bar-beep-interval"
    KEY_PROGRESS_BAR_BEEP_VERBOSITY = "progress-bar-beep-verbosity"

    def _get_setting(self, key: str, gtype: str, default: Any) -> Any:
        """Returns the dconf value for key, or default if not in dconf."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            key,
            gtype,
            default=default,
        )

    def __init__(self) -> None:
        msg = "SOUND PRESENTER: Registering D-Bus commands."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        controller = dbus_service.get_remote_controller()
        controller.register_decorated_module("SoundPresenter", self)
        self._progress_bar_cache: dict = {}

    def create_preferences_grid(
        self,
        title_change_callback: Callable[[str], None] | None = None,
    ) -> SoundPreferencesGrid:
        """Returns the GtkGrid containing the preferences UI."""

        # pylint: disable-next=import-outside-toplevel
        from .sound_presenter_preferences_grid import SoundPreferencesGrid

        return SoundPreferencesGrid(self, title_change_callback)

    @gsettings_registry.get_registry().gsetting(
        key=KEY_ENABLED,
        schema="sound",
        gtype="b",
        default=True,
        summary="Enable sound output",
        migration_key="enableSound",
    )
    @dbus_service.getter
    def get_sound_is_enabled(self) -> bool:
        """Returns whether sound is enabled."""

        return self._get_setting(self.KEY_ENABLED, "b", True)

    @dbus_service.setter
    def set_sound_is_enabled(self, value: bool) -> bool:
        """Sets whether sound is enabled."""

        msg = f"SOUND PRESENTER: Setting enable sound to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(self._SCHEMA, self.KEY_ENABLED, value)
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_VOLUME,
        schema="sound",
        gtype="d",
        default=0.5,
        summary="Sound volume (0.0-1.0)",
        migration_key="soundVolume",
    )
    @dbus_service.getter
    def get_sound_volume(self) -> float:
        """Returns the sound volume (0.0 to 1.0)."""

        return self._get_setting(self.KEY_VOLUME, "d", 0.5)

    @dbus_service.setter
    def set_sound_volume(self, value: float) -> bool:
        """Sets the sound volume (0.0 to 1.0)."""

        msg = f"SOUND PRESENTER: Setting sound volume to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(self._SCHEMA, self.KEY_VOLUME, value)
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_BEEP_PROGRESS_BAR_UPDATES,
        schema="sound",
        gtype="b",
        default=False,
        summary="Beep progress bar updates",
        migration_key="beepProgressBarUpdates",
    )
    @dbus_service.getter
    def get_beep_progress_bar_updates(self) -> bool:
        """Returns whether beep progress bar updates are enabled."""

        return self._get_setting(self.KEY_BEEP_PROGRESS_BAR_UPDATES, "b", False)

    @dbus_service.setter
    def set_beep_progress_bar_updates(self, value: bool) -> bool:
        """Sets whether beep progress bar updates are enabled."""

        msg = f"SOUND PRESENTER: Setting beep progress bar updates to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_BEEP_PROGRESS_BAR_UPDATES,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_PROGRESS_BAR_BEEP_INTERVAL,
        schema="sound",
        gtype="i",
        default=0,
        summary="Progress bar beep interval in seconds",
        migration_key="progressBarBeepInterval",
    )
    @dbus_service.getter
    def get_progress_bar_beep_interval(self) -> UInt32:
        """Returns the beep progress bar update interval in seconds."""

        return self._get_setting(self.KEY_PROGRESS_BAR_BEEP_INTERVAL, "i", 0)

    @dbus_service.setter
    def set_progress_bar_beep_interval(self, value: UInt32) -> bool:
        """Sets the beep progress bar update interval in seconds."""

        msg = f"SOUND PRESENTER: Setting progress bar beep interval to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_PROGRESS_BAR_BEEP_INTERVAL,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_PROGRESS_BAR_BEEP_VERBOSITY,
        schema="sound",
        genum="org.gnome.Orca.ProgressBarVerbosity",
        default="application",
        summary="Progress bar beep verbosity (all, application, window)",
        migration_key="progressBarBeepVerbosity",
    )
    @dbus_service.getter
    def get_progress_bar_beep_verbosity(self) -> UInt32:
        """Returns the beep progress bar verbosity level."""

        nick = gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_PROGRESS_BAR_BEEP_VERBOSITY,
            "",
            genum="org.gnome.Orca.ProgressBarVerbosity",
            default="application",
        )
        return ProgressBarVerbosity[nick.upper()].value

    @dbus_service.setter
    def set_progress_bar_beep_verbosity(self, value: UInt32) -> bool:
        """Sets the beep progress bar verbosity level."""

        msg = f"SOUND PRESENTER: Setting progress bar beep verbosity to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        level = ProgressBarVerbosity(value)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_PROGRESS_BAR_BEEP_VERBOSITY,
            level.name.lower(),
        )
        return True

    def should_present_progress_bar_update(
        self,
        obj: object,
        percent: int | None,
        is_same_app: bool,
        is_same_window: bool,
    ) -> bool:
        """Returns True if the progress bar update should be beeped."""

        if not self.get_beep_progress_bar_updates():
            return False

        last_time, last_value = self._progress_bar_cache.get(id(obj), (0.0, None))
        if percent == last_value:
            return False

        if percent != 100:
            interval = int(time.time() - last_time)
            if interval < self.get_progress_bar_beep_interval():
                return False

        verbosity = self.get_progress_bar_beep_verbosity()
        if verbosity == ProgressBarVerbosity.ALL.value:
            present = True
        elif verbosity == ProgressBarVerbosity.APPLICATION.value:
            present = is_same_app
        elif verbosity == ProgressBarVerbosity.WINDOW.value:
            present = is_same_window
        else:
            present = True

        if present:
            self._progress_bar_cache[id(obj)] = (time.time(), percent)

        return present

    def _build_generator_context(
        self,
        reason: PresentationReason | None = None,
        prior_obj: Atspi.Accessible | None = None,
    ) -> GeneratorContext:
        """Builds the settings context for sound generators."""

        mgr = focus_manager.get_manager()
        active_mode, _obj = mgr.get_active_mode_and_object_of_interest()

        return GeneratorContext(
            enabled=self.get_sound_is_enabled(),
            verbose=False,
            focus=mgr.get_locus_of_focus(),
            in_focus_mode=document_presenter.get_presenter().get_in_focus_mode(),
            active_mode=active_mode,
            reason=reason or PresentationReason.FOCUS_CHANGE,
            prior_obj=prior_obj,
            offset=None,
            leaving=False,
            ancestor_of=None,
            content_item=None,
            content_position=None,
            content_subject=None,
            resolved_role=None,
            role_subject=None,
            include_context=True,
        )

    def present_generated_sound(
        self,
        script: default.Script,
        obj: Atspi.Accessible,
        *,
        prior_obj: Atspi.Accessible | None = None,
        reason: PresentationReason | None = None,
    ) -> None:
        """Generates sound for obj using the script's sound generator and plays it."""

        context = self._build_generator_context(
            reason,
            prior_obj=prior_obj,
        )
        sounds = script.get_sound_generator().generate_sound(obj, context)
        self.play(sounds)

    def play(self, sounds: list[Icon | Tone] | Icon | Tone, interrupt: bool = True) -> None:
        """Plays the specified sound(s)."""

        if not sounds:
            return

        if not isinstance(sounds, list):
            sounds = [sounds]

        player = sound.get_player()
        player.play(sounds[0], interrupt)
        for i in range(1, len(sounds)):
            player.play(sounds[i], interrupt=False)

    def init_sound(self) -> None:
        """Initializes sound if enabled."""

        if not self.get_sound_is_enabled():
            return

        sound.get_player().init()

    def shutdown_sound(self) -> None:
        """Shuts down sound."""

        sound.get_player().shutdown()


_presenter: SoundPresenter = SoundPresenter()


def get_presenter() -> SoundPresenter:
    """Returns the Sound Presenter"""

    return _presenter



# Orca
#
# Copyright 2005-2009 Sun Microsystems Inc.
# Copyright 2011-2026 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Manager for screen reader management commands."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import (
    guilabels,
    input_event,
    preferences_window,
    screen_reader_manager_command_definitions,
    script_manager,
)
from .extension import Extension

if TYPE_CHECKING:
    from .command import Command
    from .scripts import default


class ScreenReaderManager(Extension):
    """Manages screen reader commands that control Orca itself."""

    GROUP_LABEL = guilabels.KB_GROUP_SCREEN_READER_MANAGEMENT

    def _get_commands(self) -> list[Command]:
        return screen_reader_manager_command_definitions.get_commands(self)

    def quit(
        self,
        _script: default.Script | None = None,
        _event: input_event.InputEvent | None = None,
    ) -> bool:
        """Quits Orca in response to a command."""

        from . import orca  # pylint: disable=import-outside-toplevel

        return orca.shutdown()

    def show_app_preferences_gui(
        self,
        current_script: default.Script | None = None,
        _event: input_event.InputEvent | None = None,
    ) -> bool:
        """Shows the app Preferences window."""

        current_script = current_script or script_manager.get_manager().get_active_script()
        current_script = current_script or script_manager.get_manager().get_default_script()
        if current_script is None:
            return False

        return preferences_window.show_app_preferences_gui(current_script)

    def show_preferences_gui(
        self,
        _script: default.Script | None = None,
        _event: input_event.InputEvent | None = None,
    ) -> bool:
        """Displays the Preferences window."""

        default_script = script_manager.get_manager().get_default_script()
        if default_script is None:
            return False

        script_manager.get_manager().set_active_script(default_script, "Global preferences")
        return preferences_window.show_preferences_gui(default_script)


_manager = ScreenReaderManager()


def get_manager() -> ScreenReaderManager:
    """Returns the Screen Reader Manager singleton."""

    return _manager



# Orca
#
# Copyright 2010 Joanmarie Diggs.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Custom script for Gajim."""

from __future__ import annotations

from typing import TYPE_CHECKING

from orca import chat_presenter
from orca.ax_utilities import AXUtilities
from orca.scripts import default

if TYPE_CHECKING:
    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi


class Script(default.Script):
    """Custom script for Gajim."""

    def _on_text_inserted(self, event: Atspi.Event) -> bool:
        """Callback for object:text-changed:insert accessibility events."""

        if chat_presenter.get_presenter().present_inserted_text(self, event):
            return True

        return super()._on_text_inserted(event)

    def _on_window_activated(self, event: Atspi.Event) -> bool:
        """Callback for window:activate accessibility events."""

        # Hack to "tickle" the accessible hierarchy. Otherwise, the
        # events we need to present text added to the chatroom are
        # missing.
        AXUtilities.find_all_page_tabs(event.source)
        return super()._on_window_activated(event)



# Orca
#
# Copyright 2010 Joanmarie Diggs.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.


"""Custom script utilities for pidgin."""

from __future__ import annotations

from typing import TYPE_CHECKING

from orca import script_utilities
from orca.ax_object import AXObject
from orca.ax_table import AXTable
from orca.ax_utilities import AXUtilities

if TYPE_CHECKING:
    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi


class Utilities(script_utilities.Utilities):
    """Custom script utilities for pidgin."""

    def get_expander_cell_for(self, obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the cell that is expandable in the row with obj."""

        if not self._script.chat.is_in_buddy_list(obj):
            return None

        if AXUtilities.is_expandable(obj):
            return obj

        if not AXUtilities.is_table_cell(obj):
            return None

        parent = AXObject.get_parent(obj)
        if AXUtilities.is_table_cell(parent):
            obj = parent

        candidate = AXObject.get_previous_sibling(obj)
        if AXUtilities.is_expandable(candidate):
            return candidate

        return None

    def child_nodes(self, obj: Atspi.Accessible) -> list[Atspi.Accessible]:
        """Gets all of the objects that have RELATION_NODE_CHILD_OF pointing to this cell."""

        if not self._script.chat.is_in_buddy_list(obj):
            return super().child_nodes(obj)

        if not AXUtilities.is_expanded(obj):
            return []

        parent = AXUtilities.get_table(obj)
        if parent is None:
            return []

        nodes = []
        row, col = AXTable.get_cell_coordinates(obj)

        # increment the column because the expander cell is hidden.
        col += 1
        node_level = self.node_level(obj)

        # Candidates will be in the rows beneath the current row.
        # Only check in the current column and stop checking as
        # soon as the node level of a candidate is equal or less
        # than our current level.
        #
        for i in range(row + 1, AXTable.get_row_count(parent, prefer_attribute=False)):
            cell = AXTable.get_cell_at(parent, i, col)
            node_cell = AXObject.get_previous_sibling(cell)
            targets = AXUtilities.get_is_node_child_of(node_cell)
            if not targets:
                continue

            node_of = targets[0]
            if obj == node_of:
                nodes.append(cell)
            elif self.node_level(node_of) <= node_level:
                break

        return nodes

    def node_level(self, obj: Atspi.Accessible) -> int:
        if not self._script.chat.is_in_buddy_list(obj):
            return super().node_level(obj)

        return super().node_level(AXObject.get_previous_sibling(obj))



# Orca
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Preferences grid for user extension management."""

# pylint: disable=no-member
# pylint: disable=too-many-instance-attributes

from __future__ import annotations

import os
import shutil
from typing import Any

import gi

gi.require_version("GLib", "2.0")
gi.require_version("Gtk", "3.0")
from gi.repository import GLib, Gtk  # pylint: disable=no-name-in-module

from . import (
    command_manager,
    debug,
    extension_loader,
    extension_preferences,
    extension_preferences_dialog,
    guilabels,
    orca_gui_helpers,
    preferences_grid_base,
    presentation_manager,
)


class ExtensionLoaderPreferencesGrid(preferences_grid_base.PreferencesGridBase):
    """Preferences grid for approving and disabling user extensions."""

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for user extension preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.USER_EXTENSIONS,
            summary="Use these settings to approve and enable user extensions.",
            description=guilabels.EXTENSIONS_INFO,
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.USER_EXTENSIONS,
                    kind="list",
                    summary=(
                        "Lists user extension files and actions for approving or revoking them."
                    ),
                    dynamic_values=True,
                ),
            ),
        )

    def __init__(self) -> None:
        super().__init__(guilabels.USER_EXTENSIONS)
        self._loader = extension_loader.get_loader()
        user_data_dir = GLib.get_user_data_dir()  # pylint: disable=no-value-for-parameter
        self._extensions_dir = os.path.join(user_data_dir, "orca", "extensions")
        self._listbox: Gtk.ListBox | None = None
        self._scrolled: Gtk.ScrolledWindow | None = None
        self._rows_by_filename: dict[str, Gtk.ListBoxRow] = {}
        self._switches_by_filename: dict[str, Gtk.Switch] = {}
        self._info_buttons_by_filename: dict[str, Gtk.Button] = {}
        self._settings_buttons_by_filename: dict[str, Gtk.Button] = {}
        self._delete_buttons_by_filename: dict[str, Gtk.Button] = {}
        self._approval_buttons_by_filename: dict[str, Gtk.Button] = {}
        self._summary_labels_by_filename: dict[str, Gtk.Label] = {}
        self._approval_button_size_group = orca_gui_helpers.create_horizontal_size_group()
        self._staged_settings_by_class_name: dict[
            str,
            tuple[list[extension_preferences.ExtensionPreference], dict[str, Any]],
        ] = {}
        self._syncing_filenames: set[str] = set()
        self._build()
        self.refresh()

    def _build(self) -> None:
        """Create the Gtk widgets composing the grid."""

        row = 0

        info_listbox = orca_gui_helpers.create_info_listbox(guilabels.EXTENSIONS_INFO)
        self.attach(info_listbox, 0, row, 1, 1)
        row += 1

        section_box, section_content, _heading_label = orca_gui_helpers.create_section_box(
            guilabels.USER_EXTENSIONS,
            margin_top=0,
        )
        self._scrolled = orca_gui_helpers.create_preferences_scrolled_window()
        orca_gui_helpers.add_section_content(section_box, section_content, self._scrolled)
        self.attach(section_box, 0, row, 1, 1)

    def reload(self) -> None:
        """Reload extension metadata and refresh the UI."""

        self.refresh()

    def save_settings(self, _profile: str = "", _app_name: str = "") -> dict:
        """Save settings for this grid."""

        result: dict[str, Any] = {}
        for class_name, (preferences, values) in self._staged_settings_by_class_name.items():
            extension = self._loader.get_loaded_user_extension(class_name)
            if extension is None:
                continue

            for pref in preferences:
                value = values[pref.key]
                if value == pref.default:
                    extension.settings.reset(pref.key)
                else:
                    extension.settings.set(pref.key, value)
                result[f"{class_name}/{pref.key}"] = value

        self._staged_settings_by_class_name = {}
        self._has_unsaved_changes = False
        return result

    def refresh(self) -> None:
        """Rebuild the user extension list."""

        assert self._scrolled is not None
        self._listbox = orca_gui_helpers.create_framed_listbox(
            selection_mode=Gtk.SelectionMode.SINGLE,
            accessible_name=guilabels.USER_EXTENSIONS,
        )
        self._rows_by_filename = {}
        self._switches_by_filename = {}
        self._info_buttons_by_filename = {}
        self._settings_buttons_by_filename = {}
        self._delete_buttons_by_filename = {}
        self._approval_buttons_by_filename = {}
        self._summary_labels_by_filename = {}
        self._approval_button_size_group = orca_gui_helpers.create_horizontal_size_group()

        infos = self._loader.discover_user_extensions(self._extensions_dir)
        if not infos:
            self._listbox.add(orca_gui_helpers.create_info_row(guilabels.EXTENSIONS_NO_EXTENSIONS))
        else:
            for info in infos:
                self._add_extension_row(info)

        child = self._scrolled.get_child()
        if child is not None:
            self._scrolled.remove(child)
        self._scrolled.add(self._listbox)
        self._scrolled.show_all()

    def _add_extension_row(self, info: extension_loader.UserExtensionInfo) -> None:
        """Add one extension row to the list."""

        assert self._listbox is not None

        row, summary_label, _detail_label, action_buttons = orca_gui_helpers.create_action_list_row(
            self._get_summary_text(info),
            self._get_detail_text(info),
            actions=self._get_actions(info.filename),
            stack_labels=True,
        )

        switch = orca_gui_helpers.create_switch_control(
            self._on_switch_toggled,
            info.status is extension_loader.UserExtensionStatus.APPROVED,
            "",
            handler_args=(info.filename,),
        )

        hbox = self._get_action_list_row_box(row)
        hbox.pack_start(switch, False, False, 0)

        self._listbox.add(row)
        self._rows_by_filename[info.filename] = row
        self._switches_by_filename[info.filename] = switch
        self._summary_labels_by_filename[info.filename] = summary_label
        self._info_buttons_by_filename[info.filename] = action_buttons["info"]
        self._settings_buttons_by_filename[info.filename] = action_buttons["settings"]
        self._delete_buttons_by_filename[info.filename] = action_buttons["delete"]
        self._approval_buttons_by_filename[info.filename] = action_buttons["approval"]
        self._configure_approval_button(action_buttons["approval"])
        self._sync_row(info)

    @staticmethod
    def _get_action_list_row_box(row: Gtk.ListBoxRow) -> Gtk.Box:
        """Returns the horizontal box in an action list row."""

        vbox = row.get_child()
        assert isinstance(vbox, Gtk.Box)
        children = vbox.get_children()
        assert children and isinstance(children[-1], Gtk.Box)
        return children[-1]

    def _get_actions(
        self,
        filename: str,
    ) -> list[orca_gui_helpers.ListRowAction]:
        """Returns action buttons for an extension row."""

        return [
            orca_gui_helpers.ListRowAction(
                "info",
                guilabels.EXTENSIONS_INFO_BUTTON,
                lambda _button: self._on_info_clicked(filename),
                "help-about-symbolic",
            ),
            orca_gui_helpers.ListRowAction(
                "settings",
                guilabels.EXTENSIONS_SETTINGS_BUTTON,
                lambda _button: self._on_settings_clicked(filename),
                "emblem-system-symbolic",
            ),
            orca_gui_helpers.ListRowAction(
                "delete",
                guilabels.EXTENSIONS_DELETE,
                lambda _button: self._on_delete_clicked(filename),
                "user-trash-symbolic",
            ),
            orca_gui_helpers.ListRowAction(
                "approval",
                guilabels.EXTENSIONS_APPROVE,
                lambda _button: self._on_approval_clicked(filename),
            ),
        ]

    def _configure_approval_button(self, button: Gtk.Button) -> None:
        """Configure the stable approve/revoke action button."""

        button.set_valign(Gtk.Align.CENTER)
        button.set_size_request(112, -1)
        self._approval_button_size_group.add_widget(button)

    def _get_info(self, filename: str) -> extension_loader.UserExtensionInfo | None:
        """Returns the current info for a user extension."""

        infos = self._loader.discover_user_extensions(self._extensions_dir)
        return next((info for info in infos if info.filename == filename), None)

    def _sync_row(self, info: extension_loader.UserExtensionInfo) -> None:
        """Update stable row widgets to match the extension status."""

        filename = info.filename
        summary_label = self._summary_labels_by_filename.get(filename)
        switch = self._switches_by_filename.get(filename)
        info_button = self._info_buttons_by_filename.get(filename)
        settings_button = self._settings_buttons_by_filename.get(filename)
        delete_button = self._delete_buttons_by_filename.get(filename)
        approval_button = self._approval_buttons_by_filename.get(filename)
        detail_text = self._get_detail_text(info)

        if summary_label is not None:
            orca_gui_helpers.set_stacked_label_text(
                summary_label,
                self._get_summary_text(info),
                detail_text,
            )

        status = info.status
        can_toggle = status in (
            extension_loader.UserExtensionStatus.APPROVED,
            extension_loader.UserExtensionStatus.DISABLED,
        )

        if approval_button is not None:
            self._sync_approval_button(approval_button, status)

        if info_button is not None:
            info_button.set_sensitive(True)

        if settings_button is not None:
            settings_button.set_sensitive(self._can_show_settings(info))

        if delete_button is not None:
            delete_button.set_sensitive(True)

        if switch is not None:
            self._syncing_filenames.add(filename)
            switch.set_sensitive(can_toggle)
            switch.set_active(status is extension_loader.UserExtensionStatus.APPROVED)
            self._syncing_filenames.discard(filename)

    @staticmethod
    def _sync_approval_button(
        button: Gtk.Button,
        status: extension_loader.UserExtensionStatus,
    ) -> None:
        """Sync approval action button with status."""

        label = guilabels.EXTENSIONS_APPROVE
        sensitive = status in (
            extension_loader.UserExtensionStatus.UNAPPROVED,
            extension_loader.UserExtensionStatus.MODIFIED,
            extension_loader.UserExtensionStatus.APPROVED,
            extension_loader.UserExtensionStatus.DISABLED,
        )

        if status is extension_loader.UserExtensionStatus.MODIFIED:
            label = guilabels.EXTENSIONS_REAPPROVE
        elif status in (
            extension_loader.UserExtensionStatus.APPROVED,
            extension_loader.UserExtensionStatus.DISABLED,
        ):
            label = guilabels.EXTENSIONS_REVOKE

        button.set_image(None)
        button.set_label(label)
        button.get_accessible().set_name(label)
        button.set_sensitive(sensitive)

    def _get_detail_text(self, info: extension_loader.UserExtensionInfo) -> str:
        """Returns the secondary row detail text for an extension."""

        if info.status in (
            extension_loader.UserExtensionStatus.APPROVED,
            extension_loader.UserExtensionStatus.DISABLED,
            extension_loader.UserExtensionStatus.UNAPPROVED,
        ):
            return info.description

        if info.description:
            return info.description
        status = self._get_status_label(info.status)
        return status

    @staticmethod
    def _get_display_name(info: extension_loader.UserExtensionInfo) -> str:
        """Returns the display name for an extension."""

        return info.group_label or info.class_name or guilabels.EXTENSIONS_STATUS_INVALID

    def _get_summary_text(self, info: extension_loader.UserExtensionInfo) -> str:
        """Returns the primary row text for an extension."""

        return f"{self._get_display_name(info)}: {info.filename}"

    @staticmethod
    def _get_status_label(status: extension_loader.UserExtensionStatus) -> str:
        """Returns the user-visible label for a status."""

        labels = {
            extension_loader.UserExtensionStatus.INVALID: guilabels.EXTENSIONS_STATUS_INVALID,
            extension_loader.UserExtensionStatus.UNAPPROVED: guilabels.EXTENSIONS_STATUS_UNAPPROVED,
            extension_loader.UserExtensionStatus.APPROVED: guilabels.EXTENSIONS_STATUS_APPROVED,
            extension_loader.UserExtensionStatus.DISABLED: guilabels.EXTENSIONS_STATUS_DISABLED,
            extension_loader.UserExtensionStatus.MODIFIED: guilabels.EXTENSIONS_STATUS_MODIFIED,
        }
        return labels[status]

    def _on_info_clicked(self, filename: str) -> None:
        """Show metadata for an extension."""

        info = self._get_info(filename)
        if info is None:
            return

        parent = self.get_toplevel()
        transient_for = parent if isinstance(parent, Gtk.Window) else None
        dialog = orca_gui_helpers.create_dialog(
            self._get_display_name(info),
            default_size=(640, -1),
            default_response=Gtk.ResponseType.CLOSE,
            resizable=False,
        )
        dialog.set_transient_for(transient_for)

        listbox = orca_gui_helpers.create_framed_listbox(
            selection_mode=Gtk.SelectionMode.SINGLE,
            separators=True,
        )
        listbox.set_margin_top(6)
        listbox.set_margin_bottom(6)
        listbox.set_margin_start(6)
        listbox.set_margin_end(6)

        for label, value in self._get_info_dialog_rows(info):
            if label == guilabels.EXTENSIONS_INFO_WEBSITE:
                listbox.add(orca_gui_helpers.create_key_link_row(label, value))
            else:
                listbox.add(orca_gui_helpers.create_key_value_row(label, value))

        content_area = dialog.get_content_area()
        content_area.add(listbox)

        dialog.show_all()
        dialog.run()
        dialog.destroy()

    def _can_show_settings(self, info: extension_loader.UserExtensionInfo) -> bool:
        """Returns whether info has user-visible settings."""

        if info.status is not extension_loader.UserExtensionStatus.APPROVED:
            return False
        if info.class_name is None:
            return False
        extension = self._loader.get_loaded_user_extension(info.class_name)
        if extension is None:
            return False
        return any(
            extension_preferences.is_supported_in_generated_dialog(pref)
            for pref in extension.get_preferences()
        )

    def _on_settings_clicked(self, filename: str) -> None:
        """Show settings for an extension."""

        info = self._get_info(filename)
        if info is None or info.class_name is None:
            return

        extension = self._loader.get_loaded_user_extension(info.class_name)
        if extension is None:
            return

        preferences = extension.get_preferences()
        if not preferences:
            return

        parent = self.get_toplevel()
        transient_for = parent if isinstance(parent, Gtk.Window) else None
        dialog = extension_preferences_dialog.ExtensionPreferencesDialog(
            extension,
            preferences,
            self._get_staged_settings(info.class_name),
            transient_for,
        )
        values = dialog.run()
        if values is None:
            return

        staged_preferences = [pref for pref in preferences if pref.key in values]
        self._staged_settings_by_class_name[info.class_name] = (staged_preferences, values)
        self._has_unsaved_changes = True

    def _get_staged_settings(self, class_name: str) -> dict[str, Any] | None:
        """Return currently staged settings for class_name."""

        staged = self._staged_settings_by_class_name.get(class_name)
        if staged is None:
            return None
        _preferences, values = staged
        return values

    def _get_info_dialog_rows(
        self,
        info: extension_loader.UserExtensionInfo,
    ) -> list[tuple[str, str]]:
        """Returns rows for the extension information dialog."""

        rows = [
            (guilabels.EXTENSIONS_INFO_NAME, self._get_display_name(info)),
            (guilabels.EXTENSIONS_INFO_DESCRIPTION, info.description),
            (guilabels.EXTENSIONS_INFO_LOCATION, info.filepath),
            (guilabels.EXTENSIONS_INFO_STATUS, self._get_info_status(info)),
            (guilabels.EXTENSIONS_INFO_VERSION, info.version),
            (guilabels.EXTENSIONS_INFO_AUTHOR, info.author),
            (guilabels.EXTENSIONS_INFO_ORGANIZATION, info.organization),
            (guilabels.EXTENSIONS_INFO_COPYRIGHT, info.copyright),
            (guilabels.EXTENSIONS_INFO_WEBSITE, info.website),
        ]
        return [(label, value) for label, value in rows if value]

    def _get_info_status(self, info: extension_loader.UserExtensionInfo) -> str:
        """Returns the status text for the extension information dialog."""

        status = self._get_status_label(info.status)
        conflict_note = self._get_keybinding_conflicts(info)
        if conflict_note:
            return f"{status}; {conflict_note}"
        return status

    @staticmethod
    def _get_keybinding_conflicts(info: extension_loader.UserExtensionInfo) -> str:
        """Returns any keybinding conflicts for info."""

        if info.status is not extension_loader.UserExtensionStatus.APPROVED:
            return ""

        if not info.class_name:
            return ""

        has_conflicts = command_manager.get_manager().has_user_extension_keybinding_conflicts(
            info.class_name
        )
        if not has_conflicts:
            return ""

        return guilabels.EXTENSIONS_INFO_KEYBINDING_CONFLICT

    def _on_approval_clicked(self, filename: str) -> None:
        """Approve, re-approve, or revoke approval for an extension."""

        info = self._get_info(filename)
        if info is None:
            return

        if info.status in (
            extension_loader.UserExtensionStatus.APPROVED,
            extension_loader.UserExtensionStatus.DISABLED,
        ):
            self._loader.revoke_extension(filename)
        else:
            self._loader.approve_extension_file(info.filepath)

        self._loader.reload_user_extensions(self._extensions_dir)
        if info := self._get_info(filename):
            self._sync_row(info)

    def _on_delete_clicked(self, filename: str) -> None:
        """Delete an extension file or package directory."""

        info = self._get_info(filename)
        if info is None:
            return

        if not self._confirm_delete(info):
            return

        if not self._is_deletable_extension_path(info.filepath):
            tokens = [
                "EXTENSION LOADER PREFERENCES: Refusing to delete unexpected path",
                info.filepath,
            ]
            debug.print_tokens(debug.LEVEL_WARNING, tokens, True)
            self._present_delete_error(info)
            return

        try:
            if os.path.isdir(info.filepath) and not os.path.islink(info.filepath):
                shutil.rmtree(info.filepath)
            else:
                os.remove(info.filepath)
        except OSError as error:
            msg = f"EXTENSION LOADER PREFERENCES: Could not delete {info.filepath}: {error}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            self._present_delete_error(info)
            self._loader.reload_user_extensions(self._extensions_dir)
            self.refresh()
            return

        self._loader.revoke_extension(filename)
        if info.class_name:
            disabled = [
                class_name
                for class_name in self._loader.get_disabled_extensions()
                if class_name != info.class_name
            ]
            self._loader.set_disabled_extensions(disabled)
            self._staged_settings_by_class_name.pop(info.class_name, None)

        self._loader.reload_user_extensions(self._extensions_dir)
        self.refresh()

    def _is_deletable_extension_path(self, filepath: str) -> bool:
        """Returns True if filepath is a direct child of the user extensions directory."""

        extensions_dir = os.path.realpath(self._extensions_dir)
        target = os.path.realpath(filepath)
        if os.path.dirname(target) != extensions_dir:
            return False
        return os.path.basename(target) == os.path.basename(filepath)

    def _confirm_delete(self, info: extension_loader.UserExtensionInfo) -> bool:
        """Returns True if the user confirms deleting info."""

        parent = self.get_toplevel()
        transient_for = parent if isinstance(parent, Gtk.Window) else None
        dialog = Gtk.MessageDialog(
            transient_for=transient_for,
            modal=True,
            message_type=Gtk.MessageType.WARNING,
            buttons=Gtk.ButtonsType.YES_NO,
            text=guilabels.EXTENSIONS_DELETE_CONFIRMATION_TITLE,
        )
        dialog.format_secondary_text(
            guilabels.EXTENSIONS_DELETE_CONFIRMATION_MESSAGE % self._get_display_name(info)
        )
        response = dialog.run()
        dialog.destroy()
        return response == Gtk.ResponseType.YES

    def _present_delete_error(self, info: extension_loader.UserExtensionInfo) -> None:
        """Present a deletion error message."""

        presentation_manager.get_manager().present_message(
            guilabels.EXTENSIONS_DELETE_ERROR_MESSAGE % self._get_display_name(info),
        )

    def _on_switch_toggled(
        self,
        switch: Gtk.Switch,
        _pspec: object,
        filename: str,
    ) -> None:
        """Enable or disable an extension."""

        if filename in self._syncing_filenames:
            return

        info = self._get_info(filename)
        if info is None:
            return
        if info.class_name is None:
            return

        disabled = self._loader.get_disabled_extensions()
        if switch.get_active():
            if info.class_name not in disabled:
                return
            disabled = [name for name in disabled if name != info.class_name]
        elif info.class_name not in disabled:
            disabled = [*disabled, info.class_name]
        else:
            return

        self._loader.set_disabled_extensions(disabled)
        self._loader.reload_user_extensions(self._extensions_dir)
        if info := self._get_info(filename):
            self._sync_row(info)
